import logging
import os
from datetime import datetime, timedelta

import pytest

import salt.states.file as filestate
import salt.utils.files
import salt.utils.json
import salt.utils.platform
import salt.utils.timeutil
import salt.utils.win_functions
import salt.utils.yaml
from tests.support.mock import MagicMock, PropertyMock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {filestate: {"__salt__": {}, "__opts__": {}}}


def test__tidied():
    name = os.sep + "test"
    if salt.utils.platform.is_windows():
        name = "c:" + name
    walker = [
        (os.path.join("test", "test1"), [], ["file1"]),
        (os.path.join("test", "test2", "test3"), [], []),
        (os.path.join("test", "test2"), ["test3"], ["file2"]),
        ("test", ["test1", "test2"], ["file3"]),
    ]
    today_delta = datetime.today() - salt.utils.timeutil.utcfromtimestamp(0)
    remove = MagicMock(name="file.remove")

    mystat = MagicMock()
    mystat.st_atime = today_delta.total_seconds()
    # dir = 16877
    # file = 33188
    mock_st_mode = PropertyMock(
        side_effect=[
            33188,
            33188,
            16877,
            33188,
            16877,
            16877,
            33188,
            33188,
            16877,
            33188,
            16877,
            16877,
        ]
    )
    type(mystat).st_mode = mock_st_mode
    mystat.st_size = 10

    with patch("os.walk", return_value=walker), patch(
        "os.path.islink", return_value=False
    ), patch("os.stat", return_value=mystat), patch.dict(
        filestate.__opts__, {"test": False}
    ), patch.dict(
        filestate.__salt__, {"file.remove": remove}
    ), patch(
        "os.path.isdir", return_value=True
    ):
        ret = filestate.tidied(name=name)
    exp = {
        "name": name,
        "changes": {
            "removed": [
                os.path.join("test", "test1", "file1"),
                os.path.join("test", "test2", "file2"),
                os.path.join("test", "file3"),
            ]
        },
        "result": True,
        "comment": f"Removed 3 files or directories from directory {name}",
    }
    assert ret == exp
    assert remove.call_count == 3

    remove.reset_mock()
    mock_st_mode.reset_mock()

    with patch("os.walk", return_value=walker), patch(
        "os.path.islink", return_value=False
    ), patch("os.stat", return_value=mystat), patch.dict(
        filestate.__opts__, {"test": False}
    ), patch.dict(
        filestate.__salt__, {"file.remove": remove}
    ), patch(
        "os.path.isdir", return_value=True
    ):
        ret = filestate.tidied(name=name, rmdirs=True)
    exp = {
        "name": name,
        "changes": {
            "removed": [
                os.path.join("test", "test1", "file1"),
                os.path.join("test", "test2", "file2"),
                os.path.join("test", "test2", "test3"),
                os.path.join("test", "file3"),
                os.path.join("test", "test1"),
                os.path.join("test", "test2"),
            ]
        },
        "result": True,
        "comment": f"Removed 6 files or directories from directory {name}",
    }
    assert ret == exp
    assert remove.call_count == 6


def test__bad_input():
    exp = {
        "name": "test/",
        "changes": {},
        "result": False,
        "comment": "Specified file test/ is not an absolute path",
    }
    assert filestate.tidied(name="test/") == exp
    exp = {
        "name": "/bad-directory-name/",
        "changes": {},
        "result": False,
        "comment": "/bad-directory-name/ does not exist or is not a directory.",
    }
    assert filestate.tidied(name="/bad-directory-name/") == exp


def test_tidied_with_exclude():
    name = os.sep + "test"
    if salt.utils.platform.is_windows():
        name = "c:" + name
    walker = [
        (os.path.join("test", "test1"), [], ["file1"]),
        (os.path.join("test", "test2", "test3"), [], []),
        (os.path.join("test", "test2"), ["test3"], ["file2"]),
        ("test", ["test1", "test2"], ["file3"]),
    ]
    today_delta = datetime.today() - salt.utils.timeutil.utcfromtimestamp(0)

    mystat = MagicMock()
    mystat.st_atime = today_delta.total_seconds()
    # dir = 16877
    # file = 33188
    mock_st_mode = PropertyMock(
        side_effect=[
            33188,
            33188,
            16877,
            33188,
            16877,
            16877,
            33188,
            33188,
            16877,
            33188,
            16877,
            16877,
            33188,
            33188,
            16877,
            33188,
            16877,
            16877,
        ]
    )
    type(mystat).st_mode = mock_st_mode
    mystat.st_size = 10

    remove = MagicMock(name="file.remove")

    with patch("os.walk", return_value=walker), patch(
        "os.path.islink", return_value=False
    ), patch("os.stat", return_value=mystat), patch.dict(
        filestate.__opts__, {"test": False}
    ), patch.dict(
        filestate.__salt__, {"file.remove": remove}
    ), patch(
        "os.path.isdir", return_value=True
    ):
        ret = filestate.tidied(name=name, exclude=["notfound", "file2"])
    exp = {
        "name": name,
        "changes": {
            "removed": [
                os.path.join("test", "test1", "file1"),
                os.path.join("test", "file3"),
            ]
        },
        "result": True,
        "comment": f"Removed 2 files or directories from directory {name}",
    }
    assert ret == exp
    assert remove.call_count == 2

    remove.reset_mock()
    with patch("os.walk", return_value=walker), patch(
        "os.path.islink", return_value=False
    ), patch("os.stat", return_value=mystat), patch.dict(
        filestate.__opts__, {"test": False}
    ), patch.dict(
        filestate.__salt__, {"file.remove": remove}
    ), patch(
        "os.path.isdir", return_value=True
    ):
        ret = filestate.tidied(name=name, rmdirs=True, exclude=["notfound", "file2"])
    exp = {
        "name": name,
        "changes": {
            "removed": [
                os.path.join("test", "test1", "file1"),
                os.path.join("test", "test2", "test3"),
                os.path.join("test", "file3"),
                os.path.join("test", "test1"),
                os.path.join("test", "test2"),
            ]
        },
        "result": True,
        "comment": f"Removed 5 files or directories from directory {name}",
    }
    assert ret == exp
    assert remove.call_count == 5

    remove.reset_mock()
    with patch("os.walk", return_value=walker), patch(
        "os.path.islink", return_value=False
    ), patch("os.stat", return_value=mystat), patch.dict(
        filestate.__opts__, {"test": False}
    ), patch.dict(
        filestate.__salt__, {"file.remove": remove}
    ), patch(
        "os.path.isdir", return_value=True
    ):
        ret = filestate.tidied(
            name=name,
            rmdirs=True,
            exclude=[
                "notfound",
                os.path.join("test", "test2", "file2").replace("\\", "\\\\"),
            ],
        )
    exp = {
        "name": name,
        "changes": {
            "removed": [
                os.path.join("test", "test1", "file1"),
                os.path.join("test", "test2", "file2"),
                os.path.join("test", "test2", "test3"),
                os.path.join("test", "file3"),
                os.path.join("test", "test1"),
                os.path.join("test", "test2"),
            ]
        },
        "result": True,
        "comment": f"Removed 6 files or directories from directory {name}",
    }
    assert ret == exp
    assert remove.call_count == 6


def test_tidied_with_full_path_exclude():
    name = os.sep + "test"
    if salt.utils.platform.is_windows():
        name = "c:" + name
    walker = [
        (os.path.join("test", "test1"), [], ["file1"]),
        (os.path.join("test", "test2", "test3"), [], []),
        (os.path.join("test", "test2"), ["test3"], ["file2"]),
        ("test", ["test1", "test2"], ["file3"]),
    ]
    today_delta = datetime.today() - salt.utils.timeutil.utcfromtimestamp(0)

    mystat = MagicMock()
    mystat.st_atime = today_delta.total_seconds()
    # dir = 16877
    # file = 33188
    mock_st_mode = PropertyMock(
        side_effect=[
            33188,
            33188,
            16877,
            33188,
            16877,
            16877,
            33188,
            33188,
            16877,
            33188,
            16877,
            16877,
            33188,
            33188,
            16877,
            33188,
            16877,
            16877,
        ]
    )
    type(mystat).st_mode = mock_st_mode
    mystat.st_size = 10

    remove = MagicMock(name="file.remove")

    with patch("os.walk", return_value=walker), patch(
        "os.path.islink", return_value=False
    ), patch("os.stat", return_value=mystat), patch.dict(
        filestate.__opts__, {"test": False}
    ), patch.dict(
        filestate.__salt__, {"file.remove": remove}
    ), patch(
        "os.path.isdir", return_value=True
    ):
        ret = filestate.tidied(
            name=name,
            exclude=[os.path.join("test", "test2", "file2").replace("\\", "\\\\")],
            full_path_match=True,
        )
    exp = {
        "name": name,
        "changes": {
            "removed": [
                os.path.join("test", "test1", "file1"),
                os.path.join("test", "file3"),
            ]
        },
        "result": True,
        "comment": f"Removed 2 files or directories from directory {name}",
    }
    assert ret == exp
    assert remove.call_count == 2

    remove.reset_mock()
    with patch("os.walk", return_value=walker), patch(
        "os.path.islink", return_value=False
    ), patch("os.stat", return_value=mystat), patch.dict(
        filestate.__opts__, {"test": False}
    ), patch.dict(
        filestate.__salt__, {"file.remove": remove}
    ), patch(
        "os.path.isdir", return_value=True
    ):
        ret = filestate.tidied(
            name=name,
            rmdirs=True,
            exclude=[os.path.join("test", "test2", "file2").replace("\\", "\\\\")],
            full_path_match=True,
        )
    exp = {
        "name": name,
        "changes": {
            "removed": [
                os.path.join("test", "test1", "file1"),
                os.path.join("test", "test2", "test3"),
                os.path.join("test", "file3"),
                os.path.join("test", "test1"),
                os.path.join("test", "test2"),
            ]
        },
        "result": True,
        "comment": f"Removed 5 files or directories from directory {name}",
    }
    assert ret == exp
    assert remove.call_count == 5

    remove.reset_mock()
    with patch("os.walk", return_value=walker), patch(
        "os.path.islink", return_value=False
    ), patch("os.stat", return_value=mystat), patch.dict(
        filestate.__opts__, {"test": False}
    ), patch.dict(
        filestate.__salt__, {"file.remove": remove}
    ), patch(
        "os.path.isdir", return_value=True
    ):
        ret = filestate.tidied(
            name=name,
            rmdirs=True,
            exclude=[os.path.join("test", "test2", "file2").replace("\\", "\\\\")],
            full_path_match=False,
        )
    exp = {
        "name": name,
        "changes": {
            "removed": [
                os.path.join("test", "test1", "file1"),
                os.path.join("test", "test2", "file2"),
                os.path.join("test", "test2", "test3"),
                os.path.join("test", "file3"),
                os.path.join("test", "test1"),
                os.path.join("test", "test2"),
            ]
        },
        "result": True,
        "comment": f"Removed 6 files or directories from directory {name}",
    }
    assert ret == exp
    assert remove.call_count == 6


def test_tidied_age_size_args_AND_operator_age_not_size():
    name = os.sep + "test"
    if salt.utils.platform.is_windows():
        name = "c:" + name
    walker = [
        (os.path.join("test", "test1"), [], ["file1"]),
        (os.path.join("test", "test2", "test3"), [], []),
        (os.path.join("test", "test2"), ["test3"], ["file2"]),
        ("test", ["test1", "test2"], ["file3"]),
    ]
    today_delta = (
        datetime.today() - timedelta(days=14)
    ) - salt.utils.timeutil.utcfromtimestamp(0)
    remove = MagicMock(name="file.remove")
    with patch("os.walk", return_value=walker), patch(
        "os.path.islink", return_value=False
    ), patch("os.path.getatime", return_value=today_delta.total_seconds()), patch(
        "os.path.getsize", return_value=10
    ), patch.dict(
        filestate.__opts__, {"test": False}
    ), patch.dict(
        filestate.__salt__, {"file.remove": remove}
    ), patch(
        "os.path.isdir", return_value=True
    ):
        ret = filestate.tidied(
            name=name,
            exclude=[os.path.join("test", "test2", "file2").replace("\\", "\\\\")],
            age=1,
            size=11,
            age_size_logical_operator="AND",
            age_size_only=None,
        )
    exp = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": f"Nothing to remove from directory {name}",
    }
    assert ret == exp
    assert remove.call_count == 0


def test_tidied_age_size_args_AND_operator_age_not_size_age_only():
    name = os.sep + "test"
    if salt.utils.platform.is_windows():
        name = "c:" + name
    walker = [
        (os.path.join("test", "test1"), [], ["file1"]),
        (os.path.join("test", "test2", "test3"), [], []),
        (os.path.join("test", "test2"), ["test3"], ["file2"]),
        ("test", ["test1", "test2"], ["file3"]),
    ]
    today_delta = (
        datetime.today() - timedelta(days=14)
    ) - salt.utils.timeutil.utcfromtimestamp(0)

    mystat = MagicMock()
    mystat.st_atime = today_delta.total_seconds()
    # dir = 16877
    # file = 33188
    mock_st_mode = PropertyMock(
        side_effect=[
            33188,
            33188,
            16877,
            33188,
            16877,
            16877,
        ]
    )
    type(mystat).st_mode = mock_st_mode
    mystat.st_size = 10

    remove = MagicMock(name="file.remove")

    with patch("os.walk", return_value=walker), patch(
        "os.path.islink", return_value=False
    ), patch("os.stat", return_value=mystat), patch.dict(
        filestate.__opts__, {"test": False}
    ), patch.dict(
        filestate.__salt__, {"file.remove": remove}
    ), patch(
        "os.path.isdir", return_value=True
    ):
        ret = filestate.tidied(
            name=name,
            exclude=[os.path.join("test", "test2", "file2").replace("\\", "\\\\")],
            age=1,
            size=11,
            age_size_logical_operator="AND",
            age_size_only="age",
        )
    exp = {
        "name": name,
        "changes": {
            "removed": [
                os.path.join("test", "test1", "file1"),
                os.path.join("test", "test2", "file2"),
                os.path.join("test", "file3"),
            ]
        },
        "result": True,
        "comment": f"Removed 3 files or directories from directory {name}",
    }
    assert ret == exp
    assert remove.call_count == 3


def test_tidied_age_size_args_AND_operator_size_not_age():
    name = os.sep + "test"
    if salt.utils.platform.is_windows():
        name = "c:" + name
    walker = [
        (os.path.join("test", "test1"), [], ["file1"]),
        (os.path.join("test", "test2", "test3"), [], []),
        (os.path.join("test", "test2"), ["test3"], ["file2"]),
        ("test", ["test1", "test2"], ["file3"]),
    ]
    today_delta = (
        datetime.today() - timedelta(days=14)
    ) - salt.utils.timeutil.utcfromtimestamp(0)
    remove = MagicMock(name="file.remove")
    with patch("os.walk", return_value=walker), patch(
        "os.path.islink", return_value=False
    ), patch("os.path.getatime", return_value=today_delta.total_seconds()), patch(
        "os.path.getsize", return_value=10
    ), patch.dict(
        filestate.__opts__, {"test": False}
    ), patch.dict(
        filestate.__salt__, {"file.remove": remove}
    ), patch(
        "os.path.isdir", return_value=True
    ):
        ret = filestate.tidied(
            name=name,
            exclude=[os.path.join("test", "test2", "file2").replace("\\", "\\\\")],
            age=(today_delta.days + 1),
            size=9,
            age_size_logical_operator="AND",
            age_size_only=None,
        )
    exp = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": f"Nothing to remove from directory {name}",
    }
    assert ret == exp
    assert remove.call_count == 0


def test_tidied_age_size_args_AND_operator_size_not_age_size_only():
    name = os.sep + "test"
    if salt.utils.platform.is_windows():
        name = "c:" + name
    walker = [
        (os.path.join("test", "test1"), [], ["file1"]),
        (os.path.join("test", "test2", "test3"), [], []),
        (os.path.join("test", "test2"), ["test3"], ["file2"]),
        ("test", ["test1", "test2"], ["file3"]),
    ]
    today_delta = (
        datetime.today() - timedelta(days=14)
    ) - salt.utils.timeutil.utcfromtimestamp(0)

    mystat = MagicMock()
    mystat.st_atime = today_delta.total_seconds()
    # dir = 16877
    # file = 33188
    mock_st_mode = PropertyMock(
        side_effect=[
            33188,
            33188,
            16877,
            33188,
            16877,
            16877,
        ]
    )
    type(mystat).st_mode = mock_st_mode
    mystat.st_size = 10

    remove = MagicMock(name="file.remove")

    with patch("os.walk", return_value=walker), patch(
        "os.path.islink", return_value=False
    ), patch("os.stat", return_value=mystat), patch.dict(
        filestate.__opts__, {"test": False}
    ), patch.dict(
        filestate.__salt__, {"file.remove": remove}
    ), patch(
        "os.path.isdir", return_value=True
    ):
        ret = filestate.tidied(
            name=name,
            exclude=[os.path.join("test", "test2", "file2").replace("\\", "\\\\")],
            age=(today_delta.days + 1),
            size=9,
            age_size_logical_operator="AND",
            age_size_only="size",
        )
    exp = {
        "name": name,
        "changes": {
            "removed": [
                os.path.join("test", "test1", "file1"),
                os.path.join("test", "test2", "file2"),
                os.path.join("test", "file3"),
            ]
        },
        "result": True,
        "comment": f"Removed 3 files or directories from directory {name}",
    }
    assert ret == exp
    assert remove.call_count == 3


def test_tidied_age_size_args_AND_operator_size_and_age():
    name = os.sep + "test"
    if salt.utils.platform.is_windows():
        name = "c:" + name
    walker = [
        (os.path.join("test", "test1"), [], ["file1"]),
        (os.path.join("test", "test2", "test3"), [], []),
        (os.path.join("test", "test2"), ["test3"], ["file2"]),
        ("test", ["test1", "test2"], ["file3"]),
    ]
    today_delta = (
        datetime.today() - timedelta(days=14)
    ) - salt.utils.timeutil.utcfromtimestamp(0)

    mystat = MagicMock()
    mystat.st_atime = today_delta.total_seconds()
    # dir = 16877
    # file = 33188
    mock_st_mode = PropertyMock(
        side_effect=[
            33188,
            33188,
            16877,
            33188,
            16877,
            16877,
        ]
    )
    type(mystat).st_mode = mock_st_mode
    mystat.st_size = 10

    remove = MagicMock(name="file.remove")

    with patch("os.walk", return_value=walker), patch(
        "os.path.islink", return_value=False
    ), patch("os.stat", return_value=mystat), patch.dict(
        filestate.__opts__, {"test": False}
    ), patch.dict(
        filestate.__salt__, {"file.remove": remove}
    ), patch(
        "os.path.isdir", return_value=True
    ):
        ret = filestate.tidied(
            name=name,
            exclude=[os.path.join("test", "test2", "file2").replace("\\", "\\\\")],
            age=1,
            size=9,
            age_size_logical_operator="AND",
            age_size_only=None,
        )
    exp = {
        "name": name,
        "changes": {
            "removed": [
                os.path.join("test", "test1", "file1"),
                os.path.join("test", "test2", "file2"),
                os.path.join("test", "file3"),
            ]
        },
        "result": True,
        "comment": f"Removed 3 files or directories from directory {name}",
    }
    assert ret == exp
    assert remove.call_count == 3


def test_tidied_filenotfound(tmp_path):
    name = tmp_path / "not_found_test"
    name.mkdir(parents=True, exist_ok=True)
    name = str(tmp_path / "not_found_test")
    walker = [
        (os.path.join(name, "test1"), [], ["file1"]),
        (os.path.join(name, "test2", "test3"), [], []),
        (os.path.join(name, "test2"), ["test3"], ["file2"]),
        (name, ["test1", "test2"], ["file3"]),
    ]
    # mock the walk, but files aren't there
    with patch("os.walk", return_value=walker), patch(
        "os.path.islink", return_value=False
    ):
        ret = filestate.tidied(
            name=name,
            age=1,
            size=9,
        )
    exp = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": f"Nothing to remove from directory {name}",
    }
    assert ret == exp


def test_tidied_rmlinks():
    name = os.sep + "test"
    if salt.utils.platform.is_windows():
        name = "c:" + name
    walker = [
        (os.path.join("test", "test1"), [], ["file1"]),
        (os.path.join("test", "test2", "test3"), [], []),
        (os.path.join("test", "test2"), ["test3"], ["link1"]),
        ("test", ["test1", "test2"], ["file3"]),
    ]
    today_delta = (
        datetime.today() - timedelta(days=14)
    ) - salt.utils.timeutil.utcfromtimestamp(0)

    mystat = MagicMock()
    mystat.st_atime = today_delta.total_seconds()
    # dir = 16877
    # file = 33188
    mock_st_mode = PropertyMock(
        side_effect=[
            33188,
            16877,
            33188,
            16877,
            16877,
        ]
    )
    type(mystat).st_mode = mock_st_mode
    mystat.st_size = 10

    mylstat = MagicMock()
    mylstat.st_atime = today_delta.total_seconds()
    mylstat.st_mode = 33188
    mylstat.st_size = 10

    remove = MagicMock(name="file.remove")

    with patch("os.walk", return_value=walker), patch(
        "os.path.islink", side_effect=[False, True, False, False, False, False]
    ), patch("os.lstat", return_value=mylstat), patch(
        "os.stat", return_value=mystat
    ), patch.dict(
        filestate.__opts__, {"test": False}
    ), patch.dict(
        filestate.__salt__, {"file.remove": remove}
    ), patch(
        "os.path.isdir", return_value=True
    ):
        ret = filestate.tidied(
            name=name,
            age=1,
            size=9,
            rmlinks=False,
        )
    exp = {
        "name": name,
        "changes": {
            "removed": [
                os.path.join("test", "test1", "file1"),
                os.path.join("test", "file3"),
            ]
        },
        "result": True,
        "comment": f"Removed 2 files or directories from directory {name}",
    }
    assert ret == exp
    assert remove.call_count == 2



import pytest

import salt.client.ssh.client
import salt.config
import salt.roster
import salt.utils.files
import salt.utils.path
import salt.utils.thin
import salt.utils.yaml
from salt.client import ssh
from tests.support.mock import ANY, MagicMock, patch

pytestmark = [
    pytest.mark.skipif(
        not salt.utils.path.which("ssh"), reason="No ssh binary found in path"
    ),
    pytest.mark.slow_test,
]


def test_password_failure(temp_salt_master, tmp_path):
    """
    Check password failures when trying to deploy keys
    """
    opts = temp_salt_master.config.copy()
    opts["list_hosts"] = False
    opts["argv"] = ["test.ping"]
    opts["selected_target_option"] = "glob"
    opts["tgt"] = "localhost"
    opts["arg"] = []
    roster = str(tmp_path / "roster")
    handle_ssh_ret = [
        (
            {
                "localhost": {
                    "retcode": 255,
                    "stderr": "Permission denied (publickey).\r\n",
                    "stdout": "",
                    "_error": "Permission denied",
                }
            },
            1,
        )
    ]
    key_deploy_ret = (
        {
            "localhost": {
                "retcode": 255,
                "stderr": "Permission denied (publickey)",
                "stdout": "",
                "_error": "Permission denied",
            }
        },
        1,
    )
    expected = {"localhost": "Permission denied (publickey)"}
    display_output = MagicMock()
    with patch("salt.roster.get_roster_file", MagicMock(return_value=roster)), patch(
        "salt.client.ssh.SSH.handle_ssh", MagicMock(return_value=handle_ssh_ret)
    ), patch(
        "salt.client.ssh.SSH.key_deploy", MagicMock(return_value=key_deploy_ret)
    ), patch(
        "salt.output.display_output", display_output
    ):
        client = ssh.SSH(opts)
        ret = next(client.run_iter())
        with pytest.raises(SystemExit):
            client.run()
    display_output.assert_called_once_with(expected, "nested", ANY)
    assert ret is handle_ssh_ret[0][0]



"""
    :codeauthor: Oleg Lipovchenko <oleg.lipovchenko@gmail.com>
"""

import pytest

import salt.loader
import salt.matchers.compound_match as compound_match
import salt.matchers.glob_match as glob_match
import salt.matchers.list_match as list_match
import salt.matchers.pcre_match as pcre_match
import salt.modules.match as match
from salt.exceptions import SaltException
from tests.support.mock import MagicMock, patch


@pytest.fixture
def minion_id():
    return "bar03"


@pytest.fixture
def configure_loader_modules(minion_id):
    matchers_dict = {
        "compound_match.match": compound_match.match,
        "glob_match.match": glob_match.match,
        "list_match.match": list_match.match,
        "pcre_match.match": pcre_match.match,
    }

    with patch("salt.loader.matchers", MagicMock(return_value=matchers_dict)):
        yield {
            match: {"__opts__": {"extension_modules": "", "id": minion_id}},
            compound_match: {"__opts__": {"id": minion_id}},
            glob_match: {"__opts__": {"id": minion_id}},
            list_match: {"__opts__": {"id": minion_id}},
        }


def test_compound_with_minion_id(minion_id):
    """
    Make sure that when a minion_id IS past, that it is contained in opts
    """
    mock_compound_match = MagicMock()
    target = "bar04"
    new_minion_id = "new_minion_id"

    with patch.object(
        salt.loader,
        "matchers",
        return_value={"compound_match.match": mock_compound_match},
    ) as matchers:
        match.compound(target, minion_id=new_minion_id)

        # The matcher should get called with minion_id
        matchers.assert_called_once()

        # The compound matcher should not get minion_id, no opts should be passed
        mock_compound_match.assert_called_once_with(
            target,
            minion_id="new_minion_id",
            opts={"extension_modules": "", "id": minion_id},
        )

        # Ensure that the id of the minion is bar03
        assert match.__opts__["id"] == minion_id


def test_compound(minion_id):
    """
    Test issue #55149
    """
    mock_compound_match = MagicMock()
    target = "bar04"

    with patch.object(
        salt.loader,
        "matchers",
        return_value={"compound_match.match": mock_compound_match},
    ) as matchers:
        match.compound(target)

        # The matcher should get called with minion_id
        matchers.assert_called_once()
        assert len(matchers.call_args[0]) == 1
        assert matchers.call_args[0][0].get("id") == minion_id

        # The compound matcher should not get minion_id, no opts should be passed
        mock_compound_match.assert_called_once_with(
            target, minion_id=None, opts={"extension_modules": "", "id": minion_id}
        )


def test_filter_by():
    """
    Tests if filter_by returns the correct dictionary.
    """
    lookup = {
        "foo*": {"key1": "fooval1", "key2": "fooval2"},
        "bar*": {"key1": "barval1", "key2": "barval2"},
    }
    result = {"key1": "barval1", "key2": "barval2"}

    assert match.filter_by(lookup) == result


def test_watch_for_opts_mismatch_glob_match(minion_id):
    """
    Tests for situations where the glob matcher might reference __opts__ directly
    instead of the local opts variable.

    When metaproxies/proxy minions are in use, matchers get called with a different `opts`
    dictionary.  Inside the matchers we check to see if `opts` was passed
    and use it instead of `__opts__`.  If sometime in the future we update the matchers
    and use `__opts__` directly this breaks proxy matching.
    """
    assert glob_match.match(minion_id)
    assert glob_match.match("rest03", {"id": "rest03"})
    assert not glob_match.match("rest03")


def test_watch_for_opts_mismatch_list_match(minion_id):
    """
    Tests for situations where the list matcher might reference __opts__ directly
    instead of the local opts variable

    When metaproxies/proxy minions are in use, matchers get called with a different `opts`
    dictionary.  Inside the matchers we check to see if `opts` was passed
    and use it instead of `__opts__`.  If sometime in the future we update the matchers
    and use `__opts__` directly this breaks proxy matching.
    """
    assert list_match.match(minion_id)
    assert list_match.match("rest03", {"id": "rest03"})
    assert not list_match.match("rest03")


def test_watch_for_opts_mismatch_compound_match(minion_id):
    """
    Tests for situations where the compound matcher might reference __opts__ directly
    instead of the local opts variable

    When metaproxies/proxy minions are in use, matchers get called with a different `opts`
    dictionary.  Inside the matchers we check to see if `opts` was passed
    and use it instead of `__opts__`.  If sometime in the future we update the matchers
    and use `__opts__` directly this breaks proxy matching.
    """
    assert compound_match.match(f"L@{minion_id}")
    assert compound_match.match("L@rest03", {"id": "rest03"})
    assert not compound_match.match("L@rest03")


def test_filter_by_merge():
    """
    Tests if filter_by returns a dictionary merged with another dictionary.
    """
    lookup = {
        "foo*": {"key1": "fooval1", "key2": "fooval2"},
        "bar*": {"key1": "barval1", "key2": "barval2"},
    }
    mdict = {"key1": "mergeval1"}
    result = {"key1": "mergeval1", "key2": "barval2"}

    assert match.filter_by(lookup, merge=mdict) == result


def test_filter_by_merge_lists_rep():
    """
    Tests if filter_by merges list values by replacing the original list
    values with the merged list values.
    """
    lookup = {"foo*": {"list_key": []}, "bar*": {"list_key": ["val1", "val2"]}}

    mdict = {"list_key": ["val3", "val4"]}

    # list replacement specified by the merge_lists=False option
    result = {"list_key": ["val3", "val4"]}

    assert match.filter_by(lookup, merge=mdict, merge_lists=False) == result


def test_filter_by_merge_lists_agg():
    """
    Tests if filter_by merges list values by aggregating them.
    """
    lookup = {"foo*": {"list_key": []}, "bar*": {"list_key": ["val1", "val2"]}}

    mdict = {"list_key": ["val3", "val4"]}

    # list aggregation specified by the merge_lists=True option
    result = {"list_key": ["val1", "val2", "val3", "val4"]}

    assert match.filter_by(lookup, merge=mdict, merge_lists=True) == result


def test_filter_by_merge_with_none():
    """
    Tests if filter_by merges a None object with a merge dictionary.
    """
    lookup = {"foo*": {"key1": "fooval1", "key2": "fooval2"}, "bar*": None}

    # mdict should also be the returned dictionary
    # since a merge is done with None
    mdict = {"key1": "mergeval1"}

    assert match.filter_by(lookup, merge=mdict) == mdict


def test_filter_by_merge_fail():
    """
    Tests for an exception if a merge is done without a dictionary.
    """
    lookup = {
        "foo*": {"key1": "fooval1", "key2": "fooval2"},
        "bar*": {"key1": "barval1", "key2": "barval2"},
    }
    mdict = "notadict"

    pytest.raises(SaltException, match.filter_by, lookup, merge=mdict)


def test_glob_match_different_minon_id():
    """
    Tests for situations where the glob matcher is called with a different
    minion_id than what is found in __opts__
    """
    # not passing minion_id, should return False
    assert not match.glob("bar04")

    # passing minion_id, should return True
    assert match.glob("bar04", "bar04")


def test_pcre_match_different_minion_id():
    """
    Tests for situations where the glob matcher is called with a different
    minion_id than what is found in __opts__
    """
    # not passing minion_id, should return False
    assert not match.pcre("bar.*04")

    # passing minion_id, should return True
    assert match.pcre("bar.*04", "bar04")


def test_list_match_different_minion_id():
    """
    Tests for situations where the list matcher is called with a different
    minion_id than what is found in __opts__
    """
    # not passing minion_id, should return False
    assert not match.list_("bar02,bar04")

    # passing minion_id, should return True
    assert match.list_("bar02,bar04", "bar04")


def test_ifelse():
    """
    Tests if ifelse returns the correct value.
    """
    lookup = [
        "foo*",
        {"key1": "fooval1", "key2": "fooval2"},
        "bar*",
        {"key1": "barval1", "key2": "barval2"},
    ]
    default = {"key1": "default1", "key2": "default2"}

    # even args
    with pytest.raises(SaltException):
        match.ifelse("matcher", "value")
    # only default provided
    assert match.ifelse(default, minion_id="foo03") == {
        "key1": "default1",
        "key2": "default2",
    }
    # match foo
    assert match.ifelse(*lookup, default, minion_id="foo03") == {
        "key1": "fooval1",
        "key2": "fooval2",
    }
    # match bar
    assert match.ifelse(*lookup, default, minion_id="bar03") == {
        "key1": "barval1",
        "key2": "barval2",
    }
    # no match
    assert match.ifelse(*lookup, default, minion_id="baz03") == {
        "key1": "default1",
        "key2": "default2",
    }
    # boolean matchers
    assert (
        match.ifelse(False, "nuh uhn", True, "this is true", "default value")
        == "this is true"
    )


def test_matchers_loaded_only_once(minion_id):
    """
    Test issue #62283
    """
    mock_compound_match = MagicMock()
    target = "bar04"

    with patch.object(
        salt.loader,
        "matchers",
        return_value={"compound_match.match": mock_compound_match},
    ) as matchers:
        # do this 5 times
        for run in range(0, 5):
            match.compound(target)

        # matcher loading was only performed once
        matchers.assert_called_once()
        # The matcher should get called with minion_id
        assert len(matchers.call_args[0]) == 1
        assert matchers.call_args[0][0].get("id") == minion_id

        # compound match was called 5 times
        assert mock_compound_match.call_count == 5
        # The compound matcher should not get minion_id, no opts should be passed
        mock_compound_match.assert_called_with(
            target, minion_id=None, opts={"extension_modules": "", "id": minion_id}
        )



import subprocess

import pytest
from pytestskipmarkers.utils import platform


def test_salt_call_local(salt_call_cli):
    """
    Test salt-call --local test.ping
    """
    ret = salt_call_cli.run("--local", "test.ping")
    assert ret.returncode == 0
    assert ret.data is True


def test_salt_call(salt_call_cli, salt_master):
    """
    Test salt-call test.ping
    """
    assert salt_master.is_running()

    ret = salt_call_cli.run("test.ping")
    assert ret.returncode == 0
    assert ret.data is True


@pytest.fixture
def state_name(salt_master):
    name = "some-test-state"
    sls_contents = """
    test_foo:
      test.succeed_with_changes:
          - name: foo
    """
    with salt_master.state_tree.base.temp_file(f"{name}.sls", sls_contents):
        if not platform.is_windows() and not platform.is_darwin():
            subprocess.run(
                [
                    "chown",
                    "-R",
                    "salt:salt",
                    str(salt_master.state_tree.base.write_path),
                ],
                check=False,
            )
        yield name


def test_sls(salt_call_cli, salt_master, state_name):
    """
    Test calling a sls file
    """
    assert salt_master.is_running()

    ret = salt_call_cli.run("state.apply", state_name)
    assert ret.returncode == 0
    assert ret.data
    sls_ret = ret.data[next(iter(ret.data))]
    assert sls_ret["changes"]["testing"]["new"] == "Something pretended to change"


def test_salt_call_local_sys_doc_none(salt_call_cli):
    """
    Test salt-call --local sys.doc none
    """
    ret = salt_call_cli.run("--local", "sys.doc", "none")
    assert ret.returncode == 0
    assert not ret.data


def test_salt_call_local_sys_doc_aliases(salt_call_cli):
    """
    Test salt-call --local sys.doc aliases
    """
    ret = salt_call_cli.run("--local", "sys.doc", "aliases.list_aliases")
    assert ret.returncode == 0
    assert "aliases.list_aliases" in ret.data


@pytest.mark.skip_on_windows
def test_salt_call_cmd_run_id_runas(salt_call_cli, pkg_tests_account, caplog):
    """
    Test salt-call --local cmd_run id with runas
    """
    ret = salt_call_cli.run(
        "--local", "cmd.run", "id", runas=pkg_tests_account.username
    )
    assert "Environment could not be retrieved for user" not in caplog.text
    assert str(pkg_tests_account.info.uid) in ret.stdout
    assert str(pkg_tests_account.info.gid) in ret.stdout



import logging
from collections import OrderedDict
from pathlib import Path

import pytest

import salt.loader
import salt.pillar
import salt.utils.cache
from tests.support.mock import MagicMock, patch


@pytest.mark.parametrize(
    "envs",
    (
        ["a", "b", "c"],
        ["c", "b", "a"],
        ["b", "a", "c"],
    ),
)
def test_pillar_envs_order(envs, temp_salt_minion, tmp_path):
    opts = temp_salt_minion.config.copy()
    # Stop using OrderedDict once we drop Py3.5 support
    opts["pillar_roots"] = OrderedDict()
    for envname in envs:
        opts["pillar_roots"][envname] = [str(tmp_path / envname)]
    grains = salt.loader.grains(opts)
    pillar = salt.pillar.Pillar(
        opts,
        grains,
        temp_salt_minion.id,
        "base",
    )
    # The base environment is always present and as the first environment name
    assert pillar._get_envs() == ["base"] + envs


def test_pillar_get_tops_should_not_error_when_merging_strategy_is_none_and_no_pillarenv(
    temp_salt_minion,
):
    opts = temp_salt_minion.config.copy()
    opts["pillarenv"] = None
    opts["pillar_source_merging_strategy"] = "none"
    pillar = salt.pillar.Pillar(
        opts=opts,
        grains=salt.loader.grains(opts),
        minion_id=temp_salt_minion.id,
        saltenv="base",
    )
    tops, errors = pillar.get_tops()
    assert not errors


def test_dynamic_pillarenv():
    opts = {
        "optimization_order": [0, 1, 2],
        "renderer": "json",
        "renderer_blacklist": [],
        "renderer_whitelist": [],
        "state_top": "",
        "pillar_roots": {
            "__env__": ["/srv/pillar/__env__"],
            "base": ["/srv/pillar/base"],
            "test": ["/srv/pillar/__env__"],
        },
        "file_roots": {"base": ["/srv/salt/base"], "__env__": ["/srv/salt/__env__"]},
        "extension_modules": "",
        "fileserver_backend": "roots",
        "cachedir": "",
    }
    pillar = salt.pillar.Pillar(opts, {}, "mocked-minion", "base", pillarenv="dev")
    assert pillar.opts["pillar_roots"] == {
        "base": ["/srv/pillar/base"],
        "dev": ["/srv/pillar/dev"],
        "test": ["/srv/pillar/__env__"],
    }


def test_ignored_dynamic_pillarenv():
    opts = {
        "optimization_order": [0, 1, 2],
        "renderer": "json",
        "renderer_blacklist": [],
        "renderer_whitelist": [],
        "state_top": "",
        "pillar_roots": {
            "__env__": ["/srv/pillar/__env__"],
            "base": ["/srv/pillar/base"],
        },
        "file_roots": {"base": ["/srv/salt/base"], "dev": ["/svr/salt/dev"]},
        "extension_modules": "",
        "fileserver_backend": "roots",
        "cachedir": "",
    }
    pillar = salt.pillar.Pillar(opts, {}, "mocked-minion", "base", pillarenv="base")
    assert pillar.opts["pillar_roots"] == {"base": ["/srv/pillar/base"]}


@pytest.mark.parametrize(
    "env",
    ("base", "something-else", "cool_path_123", "__env__"),
)
def test_pillar_envs_path_substitution(env, temp_salt_minion, tmp_path):
    """
    Test pillar access to a dynamic path using __env__
    """
    opts = temp_salt_minion.config.copy()

    if env == "__env__":
        # __env__ saltenv will pass "dynamic" as saltenv and
        # expect to be routed to the "dynamic" directory
        actual_env = "dynamic"
        leaf_dir = actual_env
    else:
        # any other saltenv will pass saltenv normally and
        # expect to be routed to a static "__env__" directory
        actual_env = env
        leaf_dir = "__env__"

    expected = {actual_env: [str(tmp_path / leaf_dir)]}

    # Stop using OrderedDict once we drop Py3.5 support
    opts["pillar_roots"] = OrderedDict()
    opts["pillar_roots"][env] = [str(tmp_path / leaf_dir)]
    grains = salt.loader.grains(opts)
    pillar = salt.pillar.Pillar(
        opts,
        grains,
        temp_salt_minion.id,
        actual_env,
    )

    # The __env__ string in the path has been substituted for the actual env
    assert pillar.opts["pillar_roots"] == expected


def test_pillar_get_cache_disk(temp_salt_minion, caplog):
    # create faked path for cache
    with pytest.helpers.temp_directory() as temp_path:
        opts = temp_salt_minion.config.copy()
        tmp_cachefile = Path(str(temp_path)) / "pillar" / f"{temp_salt_minion.id}.p"
        tmp_cachefile.parent.mkdir(parents=True)
        tmp_cachefile.touch()
        assert tmp_cachefile.exists()

        opts["pillarenv"] = None
        opts["pillar_cache"] = True
        opts["minion_data_cache"] = False
        opts["cachedir"] = str(temp_path)

        caplog.at_level(logging.DEBUG)
        pillar = salt.pillar.PillarCache(
            opts=opts,
            grains=salt.loader.grains(opts),
            minion_id=temp_salt_minion.id,
            saltenv="base",
        )
        fresh_pillar = pillar.compile_pillar()
        assert "Unpack failed: incomplete input" not in caplog.messages
        assert fresh_pillar == {}


def test_pillar_fetch_pillar_override_skipped(temp_salt_minion, caplog):
    with pytest.helpers.temp_directory() as temp_path:
        opts = temp_salt_minion.config.copy()
        tmp_cachefile = Path(str(temp_path)) / "pillar" / f"{temp_salt_minion.id}.p"
        tmp_cachefile.parent.mkdir(parents=True)
        tmp_cachefile.touch()
        assert tmp_cachefile.exists()

        opts = temp_salt_minion.config.copy()
        opts["pillarenv"] = None
        opts["pillar_cache"] = True
        opts["minion_data_cache"] = False
        opts["cachedir"] = str(temp_path)

        pillar_override = {"inline_pillar": True}

        caplog.at_level(logging.DEBUG)
        pillar = salt.pillar.PillarCache(
            opts=opts,
            grains=salt.loader.grains(opts),
            minion_id=temp_salt_minion.id,
            saltenv="base",
            pillar_override=pillar_override,
        )

        fresh_pillar = pillar.compile_pillar()
        assert "inline_pillar" in fresh_pillar

        pillar_cache = pillar.cache.fetch("pillar", f"{temp_salt_minion.id}:base")
        assert "inline_pillar" not in pillar_cache


def test_remote_pillar_timeout(temp_salt_minion, tmp_path):
    opts = temp_salt_minion.config.copy()
    opts["master_uri"] = "tcp://127.0.0.1:12323"
    grains = salt.loader.grains(opts)
    pillar = salt.pillar.RemotePillar(
        opts,
        grains,
        temp_salt_minion.id,
        "base",
    )
    mock = MagicMock()
    mock.side_effect = salt.exceptions.SaltReqTimeoutError()
    pillar.channel.crypted_transfer_decode_dictentry = mock
    msg = r"^Pillar timed out after \d{1,4} seconds$"
    with pytest.raises(salt.exceptions.SaltClientError):
        pillar.compile_pillar()


def test_ext_pillar_dunder_in_modules_in_pillar(temp_salt_minion):
    """
    Test that ext_pillar is available in __pillar__ inside execution modules
    during pillar render when using ext_pillar_first=true
    """
    opts = temp_salt_minion.config.copy()
    opts["ext_pillar_first"] = True

    grains = salt.loader.grains(opts)
    pillar = salt.pillar.Pillar(opts, grains, temp_salt_minion.id, "base")
    # Pillar should be empty to start with
    assert pillar.functions.pack["__pillar__"] == {}

    ext_value = {"ext": "some ext value"}
    pil_value = {"pillar": "some pillar value"}
    with patch.object(pillar, "ext_pillar", return_value=(ext_value, [])):
        with patch.object(pillar, "render_pillar", return_value=(pil_value, [])):
            compiled = pillar.compile_pillar()
            assert compiled == dict(**ext_value, **pil_value)

    # Loader should pack the opts pillar dict from the ext_pillar() call
    # and the rendered pillar data
    assert pillar.functions.pack["__pillar__"] == dict(**ext_value, **pil_value)

    # Ensure a module function can access the pillar data
    assert pillar.functions["pillar.get"]("ext") == "some ext value"
    assert pillar.functions["pillar.get"]("pillar") == "some pillar value"


def test_pillar_opts_in_dunder_pillar(temp_salt_minion):
    """
    Test that pillar_opts=True correctly includes master config in __pillar__
    during the pillar rendering process.
    """
    opts = temp_salt_minion.config.copy()
    opts["pillar_opts"] = True
    # Ensure some master config is present
    opts["master_key"] = "master_secret"

    grains = salt.loader.grains(opts)
    pillar = salt.pillar.Pillar(opts, grains, temp_salt_minion.id, "base")

    pil_value = {"pillar_key": "pillar_value"}
    with patch.object(pillar, "render_pillar", return_value=(pil_value, [])):
        compiled = pillar.compile_pillar()
        # Compiled pillar should include master opts nested under "master" when pillar_opts=True
        assert "master" in compiled
        assert compiled["master"]["master_key"] == "master_secret"
        assert compiled["pillar_key"] == "pillar_value"

    # The loader pack should also contain the master opts
    assert "master" in pillar.functions.pack["__pillar__"]
    assert (
        pillar.functions.pack["__pillar__"]["master"]["master_key"] == "master_secret"
    )
    assert pillar.functions["pillar.get"]("master:master_key") == "master_secret"


def test_ssh_merge_pillar_in_dunder_pillar(temp_salt_minion):
    """
    Test that ssh_merge_pillar correctly merges extra pillar data from opts["pillar"] into __pillar__
    """
    opts = temp_salt_minion.config.copy()
    opts["ssh_merge_pillar"] = True
    opts["pillar"] = {"ssh_key": "ssh_value"}

    grains = salt.loader.grains(opts)
    pillar = salt.pillar.Pillar(opts, grains, temp_salt_minion.id, "base")

    pil_value = {"normal_key": "normal_value"}
    with patch.object(pillar, "render_pillar", return_value=(pil_value, [])):
        compiled = pillar.compile_pillar()
        assert compiled["ssh_key"] == "ssh_value"
        assert compiled["normal_key"] == "normal_value"

    # The loader pack should contain the merged SSH pillar
    assert pillar.functions.pack["__pillar__"]["ssh_key"] == "ssh_value"
    assert pillar.functions["pillar.get"]("ssh_key") == "ssh_value"


def test_decrypt_pillar_in_dunder_pillar(temp_salt_minion):
    """
    Test that __pillar__ contains decrypted values if decryption is performed.
    """
    opts = temp_salt_minion.config.copy()

    grains = salt.loader.grains(opts)
    pillar = salt.pillar.Pillar(opts, grains, temp_salt_minion.id, "base")

    # Mock encrypted and decrypted data
    encrypted_pil = {"secret": "encrypted_value"}
    decrypted_pil = {"secret": "decrypted_value"}

    with patch.object(pillar, "render_pillar", return_value=(encrypted_pil, [])):
        # Mock the decrypt_pillar method of the Pillar class
        with patch.object(
            pillar,
            "decrypt_pillar",
            side_effect=lambda p: p.update(decrypted_pil) or [],
        ):
            compiled = pillar.compile_pillar()
            assert compiled["secret"] == "decrypted_value"

    # The loader pack should contain the decrypted pillar
    assert pillar.functions.pack["__pillar__"]["secret"] == "decrypted_value"
    assert pillar.functions["pillar.get"]("secret") == "decrypted_value"


def test_ext_pillar_after_dunder_pillar(temp_salt_minion):
    """
    Test that ext_pillar is correctly merged into __pillar__ when ext_pillar_first=False
    """
    opts = temp_salt_minion.config.copy()
    opts["ext_pillar_first"] = False

    grains = salt.loader.grains(opts)
    pillar = salt.pillar.Pillar(opts, grains, temp_salt_minion.id, "base")

    pil_value = {"normal": "value"}
    ext_value = {"ext": "value"}

    with patch.object(pillar, "render_pillar", return_value=(pil_value, [])):
        with patch.object(pillar, "ext_pillar", return_value=(ext_value, [])):
            compiled = pillar.compile_pillar()
            # Verify both are present in the final result
            assert compiled["normal"] == "value"
            assert compiled["ext"] == "value"

    # Loader should have both
    assert pillar.functions.pack["__pillar__"]["normal"] == "value"
    assert pillar.functions.pack["__pillar__"]["ext"] == "value"
    assert pillar.functions["pillar.get"]("ext") == "value"
    assert pillar.functions["pillar.get"]("normal") == "value"



"""
Integration tests for salt-ssh py_versions
"""

import logging
import os
import signal
import subprocess
import sys
import tempfile
import time

import pytest
from pytestshellutils.utils.processes import ProcessResult, terminate_process
from saltfactories.utils import random_string

from salt.utils.versions import Version
from tests.support.helpers import Keys

docker = pytest.importorskip("docker")


log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("dockerd"),
    pytest.mark.skipif(
        Version(docker.__version__) < Version("4.0.0"),
        reason="Test does not work in this version of docker-py",
    ),
]


@pytest.fixture(scope="module")
def ssh_keys(tmp_path_factory):
    """
    Temporary ssh key fixture
    """
    with Keys(tmp_path_factory) as keys:
        yield keys


@pytest.fixture(scope="module")
def ssh_password():
    return random_string("sshpassword")


@pytest.fixture(scope="module")
def ssh_container_name():
    return random_string("ssh-container-")


@pytest.fixture(scope="module")
def ssh_sub_container_name():
    return random_string("ssh-sub-container-")


@pytest.fixture(scope="module")
def ssh_container(salt_factories, ssh_container_name, ssh_password):
    """
    Temporary docker container with python 3.6 and ssh enabled
    """
    container = salt_factories.get_container(
        ssh_container_name,
        "dwoz1/cicd:ssh",
        container_run_kwargs={
            "ports": {
                "22/tcp": None,
            },
            "environment": {
                "SSH_USER": "centos",
                "SSH_USER_PASSWORD": ssh_password,
                "SSH_PASSWORD_AUTHENTICATION": "true",
            },
            "cap_add": ["IPC_LOCK"],
        },
        pull_before_start=True,
        skip_on_pull_failure=True,
        skip_if_docker_client_not_connectable=True,
    )
    with container.started() as factory:
        yield factory


@pytest.fixture(scope="module")
def ssh_sub_container(salt_factories, ssh_sub_container_name, ssh_password):
    """
    Temporary docker container with python 3.6 and ssh enabled
    """
    container = salt_factories.get_container(
        ssh_sub_container_name,
        "dwoz1/cicd:ssh",
        container_run_kwargs={
            "ports": {
                "22/tcp": None,
            },
            "environment": {
                "SSH_USER": "centos",
                "SSH_USER_PASSWORD": ssh_password,
                "SSH_PASSWORD_AUTHENTICATION": "true",
            },
            "cap_add": ["IPC_LOCK"],
        },
        pull_before_start=True,
        skip_on_pull_failure=True,
        skip_if_docker_client_not_connectable=True,
    )
    with container.started() as factory:
        yield factory


@pytest.fixture(scope="module")
def ssh_port(ssh_container):
    return ssh_container.get_host_port_binding(22, protocol="tcp")


@pytest.fixture(scope="module")
def ssh_sub_port(ssh_sub_container):
    return ssh_sub_container.get_host_port_binding(22, protocol="tcp")


@pytest.fixture(scope="module")
def salt_ssh_roster_file(
    ssh_container_name,
    ssh_sub_container_name,
    ssh_port,
    ssh_sub_port,
    salt_master,
    known_hosts_file,
):
    """
    Temporary roster for ssh docker container
    """
    roster = f"""
    {ssh_container_name}:
      host: localhost
      user: centos
      port: {ssh_port}
      ssh_options:
        - UserKnownHostsFile={known_hosts_file}
    {ssh_sub_container_name}:
      host: localhost
      user: centos
      port: {ssh_sub_port}
      ssh_options:
        - UserKnownHostsFile={known_hosts_file}
    """
    with pytest.helpers.temp_file(
        "setup_roster", roster, salt_master.config_dir
    ) as roster_file:
        yield roster_file


@pytest.fixture(scope="module")
def salt_ssh_cli(
    salt_master, salt_ssh_roster_file, ssh_keys, ssh_container, ssh_sub_container
):
    assert salt_master.is_running()
    assert ssh_container.is_running()
    assert ssh_sub_container.is_running()
    return salt_master.salt_ssh_cli(
        timeout=180,
        roster_file=salt_ssh_roster_file,
        client_key=str(ssh_keys.priv_path),
    )


@pytest.mark.flaky_jail
def test_setup(salt_ssh_cli, ssh_container_name, ssh_sub_container_name, ssh_password):
    """
    Test salt-ssh setup works
    """
    # Provide the passwd from the CLI to allow the key deploy
    possible_ids = (ssh_container_name, ssh_sub_container_name)
    ret = salt_ssh_cli.run(
        "--passwd", ssh_password, "--key-deploy", "test.ping", minion_tgt="*"
    )
    try:
        assert ret.returncode == 0
    except AssertionError:
        # Sleep and Repeat in case of failure to reduce flakyness
        time.sleep(5)
        ret = salt_ssh_cli.run(
            "--passwd", ssh_password, "--key-deploy", "test.ping", minion_tgt="*"
        )
        assert ret.returncode == 0
    for id in possible_ids:
        assert id in ret.data
        assert ret.data[id] is True

    # Run it again without the key deploy
    ret = salt_ssh_cli.run("test.ping", minion_tgt="*")
    assert ret.returncode == 0
    for id in possible_ids:
        assert id in ret.data
        assert ret.data[id] is True

    # Run a test.sleep and kill it
    sleep_time = 15
    cmdline = salt_ssh_cli.cmdline("test.sleep", sleep_time, minion_tgt="*")
    terminal_stdout = tempfile.SpooledTemporaryFile(512000, buffering=0)
    terminal_stderr = tempfile.SpooledTemporaryFile(512000, buffering=0)

    proc = subprocess.Popen(
        cmdline,
        shell=False,
        stdout=terminal_stdout,
        stderr=terminal_stderr,
        universal_newlines=True,
    )
    start = time.time()
    try:
        # Make sure it actually starts
        proc.wait(1)
    except subprocess.TimeoutExpired:
        pass
    else:
        terminate_process(proc.pid, kill_children=True)
        pytest.fail("The test process failed to start")

    time.sleep(2)
    # Send CTRL-C to the process
    os.kill(proc.pid, signal.SIGINT)
    with proc:
        # Wait for the process to terminate, to avoid zombies.
        # Shouldn't really take the 30 seconds
        proc.wait(sleep_time * 2)
        # poll the terminal so the right returncode is set on the popen object
        proc.poll()
        # This call shouldn't really be necessary
        proc.communicate()
    stop = time.time()

    terminal_stdout.flush()
    terminal_stdout.seek(0)
    stdout = proc._translate_newlines(
        terminal_stdout.read(), __salt_system_encoding__, sys.stdout.errors
    )
    terminal_stdout.close()

    terminal_stderr.flush()
    terminal_stderr.seek(0)
    stderr = proc._translate_newlines(
        terminal_stderr.read(), __salt_system_encoding__, sys.stderr.errors
    )
    terminal_stderr.close()
    ret = ProcessResult(
        returncode=proc.returncode, stdout=stdout, stderr=stderr, cmdline=proc.args
    )
    log.debug(ret)
    # If the minion ID is on stdout it means that the command finished and wasn't terminated
    for id in possible_ids:
        assert (
            id not in ret.stdout
        ), "The command wasn't actually terminated. Took {} seconds.".format(
            round(stop - start, 2)
        )



"""
tests for pkg state
"""

import logging
import os
import subprocess
import time

import pytest

import salt.utils.files
import salt.utils.path
import salt.utils.pkg.rpm
import salt.utils.platform

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_not_root,
    pytest.mark.destructive_test,
    pytest.mark.windows_whitelisted,
    pytest.mark.timeout_unless_on_windows(240),
    pytest.mark.skipif(
        bool(salt.utils.path.which("transactional-update")),
        reason="Skipping on transactional systems",
    ),
]


@pytest.fixture(scope="module", autouse=True)
def refresh_db(grains, modules):

    if salt.utils.platform.is_windows():
        modules.winrepo.update_git_repos()

    modules.pkg.refresh_db()

    # If this is Arch Linux, check if pacman is in use by another process
    if grains["os_family"] == "Arch":
        for _ in range(12):
            if not os.path.isfile("/var/lib/pacman/db.lck"):
                break
            else:
                time.sleep(5)
        else:
            pytest.fail("Package database locked after 60 seconds, bailing out")


@pytest.fixture(scope="module", autouse=True)
def refresh_keys(grains, modules):
    if grains["os_family"] == "Arch":
        # We should be running this periodically when building new test runner
        # images, otherwise this could take several minutes to complete.
        proc = subprocess.run(["pacman-key", "--refresh-keys"], check=False)
        if proc.returncode != 0:
            pytest.fail("pacman-key --refresh-keys command failed.")


@pytest.fixture
def PKG_TARGETS(grains):
    _PKG_TARGETS = ["figlet", "sl"]
    if grains["os"] == "Windows":
        _PKG_TARGETS = ["npp_x64", "putty"]
    elif grains["os"] == "Amazon":
        if grains["osfinger"] == "Amazon Linux-2023":
            _PKG_TARGETS = ["lynx", "gnuplot-minimal"]
        else:
            _PKG_TARGETS = ["lynx", "gnuplot"]
    elif grains["os_family"] == "RedHat":
        if grains["os"] == "VMware Photon OS":
            _PKG_TARGETS = ["zsh", "pciutils"]
        elif (
            grains["os"] in ("CentOS Stream", "Rocky", "AlmaLinux")
            and grains["osmajorrelease"] >= 9
        ):
            _PKG_TARGETS = ["units", "zsh"]
        else:
            _PKG_TARGETS = ["units", "zsh-html"]
    elif grains["os_family"] == "Suse":
        _PKG_TARGETS = ["iotop", "screen"]
    return _PKG_TARGETS


@pytest.fixture
def PKG_CAP_TARGETS(grains):
    _PKG_CAP_TARGETS = []
    if grains["os_family"] == "Suse":
        if grains["os"] == "SUSE":
            _PKG_CAP_TARGETS = [("perl(YAML)", "perl-YAML")]
            # sudo zypper install 'perl(YAML)'
            # Loading repository data...
            # Reading installed packages...
            # 'perl(YAML)' not found in package names. Trying capabilities.
            # Resolving package dependencies...
            #
            # The following NEW package is going to be installed:
            #   perl-YAML
            #
            # 1 new package to install.
            # Overall download size: 85.3 KiB. Already cached: 0 B. After the operation, additional 183.3 KiB will be used.
            # Continue? [y/n/v/...? shows all options] (y):

            # So, it just doesn't work here? skip it for now
            _PKG_CAP_TARGETS.clear()
    if not _PKG_CAP_TARGETS:
        pytest.skip("Capability not provided")
    return _PKG_CAP_TARGETS


@pytest.fixture
def PKG_32_TARGETS(grains):
    _PKG_32_TARGETS = []
    if grains["os_family"] == "RedHat" and grains["oscodename"] != "Photon":
        if grains["os"] == "CentOS":
            if grains["osmajorrelease"] == 5:
                _PKG_32_TARGETS = ["xz-devel.i386"]
            else:
                _PKG_32_TARGETS.append("xz-devel.i686")
    elif grains["os"] == "Windows":
        # Prefer putty first: 32-bit ``npp`` winrepo metadata and uninstall
        # registry cleanup have been flaky on Windows Server 2022/2025 CI,
        # causing pkg.removed / latest_version to disagree with reality.
        _PKG_32_TARGETS = ["putty", "npp"]
    if not _PKG_32_TARGETS:
        pytest.skip("No 32 bit packages have been specified for testing")
    return _PKG_32_TARGETS


@pytest.fixture
def PKG_DOT_TARGETS(grains):
    _PKG_DOT_TARGETS = []
    if grains["os_family"] == "RedHat" and grains["oscodename"] != "Photon":
        if grains["osmajorrelease"] == 7:
            _PKG_DOT_TARGETS = ["tomcat-el-2.2-api"]
        elif grains["osmajorrelease"] == 8:
            _PKG_DOT_TARGETS = ["aspnetcore-runtime-6.0"]
    if not _PKG_DOT_TARGETS:
        pytest.skip(
            'No packages with "." in their name have been specified',
        )
    return _PKG_DOT_TARGETS


@pytest.fixture
def PKG_EPOCH_TARGETS(grains):
    _PKG_EPOCH_TARGETS = []
    if grains["os_family"] == "RedHat" and grains["oscodename"] != "Photon":
        if grains["osmajorrelease"] == 7:
            _PKG_EPOCH_TARGETS = ["comps-extras"]
        elif grains["osmajorrelease"] == 8:
            _PKG_EPOCH_TARGETS = ["traceroute"]
    if not _PKG_EPOCH_TARGETS:
        pytest.skip('No targets have been configured with "epoch" in the version')
    return _PKG_EPOCH_TARGETS


@pytest.fixture
def VERSION_SPEC_SUPPORTED(grains):
    _VERSION_SPEC_SUPPORTED = True
    if grains["os"] == "FreeBSD":
        _VERSION_SPEC_SUPPORTED = False
    if not _VERSION_SPEC_SUPPORTED:
        pytest.skip("Version specification not supported")
    return _VERSION_SPEC_SUPPORTED


@pytest.fixture
def WILDCARDS_SUPPORTED(grains):
    _WILDCARDS_SUPPORTED = False
    if grains["os_family"] in ("Arch", "Debian"):
        _WILDCARDS_SUPPORTED = True
    if not _WILDCARDS_SUPPORTED:
        pytest.skip("Wildcards in pkg.install are not supported")
    return _WILDCARDS_SUPPORTED


@pytest.fixture
def ctx():
    return {}


@pytest.fixture
def latest_version(ctx, modules):
    """
    Helper function which ensures that we don't make any unnecessary calls to
    pkg.latest_version to figure out what version we need to install. This
    won't stop pkg.latest_version from being run in a pkg.latest state, but it
    will reduce the amount of times we check the latest version here in the
    test suite.
    """

    def run_command(*names):
        key = "latest_version"
        if key not in ctx:
            ctx[key] = dict()
        targets = [x for x in names if x not in ctx[key]]
        if targets:
            result = modules.pkg.latest_version(*targets, refresh=False)
            try:
                ctx[key].update(result)
            except ValueError:
                # Only a single target, pkg.latest_version returned a string
                ctx[key][targets[0]] = result

        ret = {x: ctx[key].get(x, "") for x in names}
        if len(names) == 1:
            return ret[names[0]]
        return ret

    return run_command


@pytest.fixture(scope="function")
def install_7zip(modules):
    try:
        modules.pkg.install(name="7zip", version="22.01.00.0")
        modules.pkg.install(name="7zip", version="19.00.00.0")
        versions = modules.pkg.version("7zip")
        assert "19.00.00.0" in versions
        assert "22.01.00.0" in versions
        yield
    finally:
        modules.pkg.remove(name="7zip", version="19.00.00.0")
        modules.pkg.remove(name="7zip", version="22.01.00.0")
        versions = modules.pkg.version("7zip")
        assert "19.00.00.0" not in versions
        assert "22.01.00.0" not in versions


@pytest.fixture(scope="module")
def pkg_def_contents(state_tree):
    return r"""
    my-software:
      '1.0.1':
        full_name: 'My Software'
        installer: 'C:\files\mysoftware101.msi'
        install_flags: '/qn /norestart'
        uninstaller: 'C:\files\mysoftware101.msi'
        uninstall_flags: '/qn /norestart'
        msiexec: True
        reboot: False
      '1.0.2':
        full_name: 'My Software'
        installer: 'C:\files\mysoftware102.msi'
        install_flags: '/qn /norestart'
        uninstaller: 'C:\files\mysoftware102.msi'
        uninstall_flags: '/qn /norestart'
        msiexec: True
        reboot: False
    """


@pytest.mark.requires_salt_modules("pkg.version")
@pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
@pytest.mark.slow_test
def test_pkg_001_installed(modules, states, PKG_TARGETS):
    """
    This is a destructive test as it installs and then removes a package
    """
    target = PKG_TARGETS[0]
    version = modules.pkg.version(target)

    # If this assert fails, we need to find new targets, this test needs to
    # be able to test successful installation of packages, so this package
    # needs to not be installed before we run the states below
    assert not version

    ret = states.pkg.installed(name=target, refresh=False)
    assert ret.result is True
    ret = states.pkg.removed(name=target)
    assert ret.result is True


@pytest.mark.usefixtures("VERSION_SPEC_SUPPORTED")
@pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
@pytest.mark.slow_test
def test_pkg_002_installed_with_version(PKG_TARGETS, states, latest_version):
    """
    This is a destructive test as it installs and then removes a package
    """
    target = PKG_TARGETS[0]
    version = latest_version(target)

    # If this assert fails, we need to find new targets, this test needs to
    # be able to test successful installation of packages, so this package
    # needs to not be installed before we run the states below
    assert version

    ret = states.pkg.installed(name=target, version=version, refresh=False)
    assert ret.result is True
    ret = states.pkg.removed(name=target)
    assert ret.result is True


@pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
@pytest.mark.slow_test
def test_pkg_003_installed_multipkg(caplog, PKG_TARGETS, modules, states, grains):
    """
    This is a destructive test as it installs and then removes two packages
    """
    if grains["os_family"] == "Arch":
        pytest.skip("Arch needs refresh_db logic added to golden image")

    version = modules.pkg.version(*PKG_TARGETS)

    # If this assert fails, we need to find new targets, this test needs to
    # be able to test successful installation of packages, so these
    # packages need to not be installed before we run the states below
    assert not any(version.values())
    ret = states.pkg.removed(name=None, pkgs=PKG_TARGETS)
    assert ret.result is True

    try:
        ret = states.pkg.installed(name=None, pkgs=PKG_TARGETS, refresh=False)
        assert ret.result is True
        if not salt.utils.platform.is_windows():
            assert "WARNING" not in caplog.text
    finally:
        ret = states.pkg.removed(name=None, pkgs=PKG_TARGETS)
        assert ret.result is True


@pytest.mark.usefixtures("VERSION_SPEC_SUPPORTED")
@pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
@pytest.mark.slow_test
def test_pkg_004_installed_multipkg_with_version(
    PKG_TARGETS, latest_version, states, grains
):
    """
    This is a destructive test as it installs and then removes two packages
    """
    if grains["os_family"] == "Arch":
        pytest.skip("Arch needs refresh_db logic added to golden image")
    version = latest_version(PKG_TARGETS[0])

    # If this assert fails, we need to find new targets, this test needs to
    # be able to test successful installation of packages, so these
    # packages need to not be installed before we run the states below
    assert bool(version)

    pkgs = [{PKG_TARGETS[0]: version}, PKG_TARGETS[1]]

    try:
        ret = states.pkg.installed(name=None, pkgs=pkgs, refresh=False)
        assert ret.result is True
    finally:
        ret = states.pkg.removed(name=None, pkgs=PKG_TARGETS)
        assert ret.result is True


@pytest.mark.requires_salt_modules("pkg.version")
@pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
@pytest.mark.slow_test
def test_pkg_005_installed_32bit(PKG_32_TARGETS, modules, states):
    """
    This is a destructive test as it installs and then removes a package
    """
    target = PKG_32_TARGETS[0]

    # _PKG_TARGETS_32 is only populated for platforms for which Salt has to
    # munge package names for 32-bit-on-x86_64 (Currently only Ubuntu and
    # RHEL-based). Don't actually perform this test on other platforms.
    version = modules.pkg.version(target)

    # If this assert fails, we need to find a new target. This test
    # needs to be able to test successful installation of packages, so
    # the target needs to not be installed before we run the states
    # below
    assert not version

    ret = states.pkg.installed(name=target, refresh=False)
    assert ret.result is True
    ret = states.pkg.removed(name=target)
    assert ret.result is True


@pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
@pytest.mark.slow_test
def test_pkg_006_installed_32bit_with_version(PKG_32_TARGETS, latest_version, states):
    """
    This is a destructive test as it installs and then removes a package
    """
    target = PKG_32_TARGETS[0]

    # _PKG_TARGETS_32 is only populated for platforms for which Salt has to
    # munge package names for 32-bit-on-x86_64 (Currently only Ubuntu and
    # RHEL-based). Don't actually perform this test on other platforms.
    version = latest_version(target)

    # If this assert fails, we need to find a new target. This test
    # needs to be able to test successful installation of the package, so
    # the target needs to not be installed before we run the states
    # below
    assert version

    ret = states.pkg.installed(name=target, version=version, refresh=False)
    assert ret.result is True
    ret = states.pkg.removed(name=target)
    assert ret.result is True


@pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
@pytest.mark.slow_test
def test_pkg_007_with_dot_in_pkgname(PKG_DOT_TARGETS, latest_version, states):
    """
    This tests for the regression found in the following issue:
    https://github.com/saltstack/salt/issues/8614

    This is a destructive test as it installs a package
    """
    target = PKG_DOT_TARGETS[0]

    version = latest_version(target)
    # If this assert fails, we need to find a new target. This test
    # needs to be able to test successful installation of the package, so
    # the target needs to not be installed before we run the
    # pkg.installed state below
    assert bool(version)
    ret = states.pkg.installed(name=target, refresh=False)
    assert ret.result is True
    ret = states.pkg.removed(name=target)
    assert ret.result is True


@pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
@pytest.mark.slow_test
def test_pkg_008_epoch_in_version(PKG_EPOCH_TARGETS, latest_version, states):
    """
    This tests for the regression found in the following issue:
    https://github.com/saltstack/salt/issues/8614

    This is a destructive test as it installs a package
    """
    target = PKG_EPOCH_TARGETS[0]

    version = latest_version(target)
    # If this assert fails, we need to find a new target. This test
    # needs to be able to test successful installation of the package, so
    # the target needs to not be installed before we run the
    # pkg.installed state below
    assert version
    ret = states.pkg.installed(name=target, version=version, refresh=False)
    assert ret.result is True
    ret = states.pkg.removed(name=target)
    assert ret.result is True


@pytest.mark.requires_salt_modules("pkg.version", "pkg.info_installed")
@pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
@pytest.mark.slow_test
def test_pkg_009_latest_with_epoch(grains, modules, states):
    """
    This tests for the following issue:
    https://github.com/saltstack/salt/issues/31014

    This is a destructive test as it installs a package
    """

    if grains["kernel"] != "linux":
        pytest.skip("Only runs on Linux.")

    if grains["os"] != "Amazon":
        pytest.skip("Does not runs on Amazon Linux.")

    package = "bash-completion"
    pkgquery = "version"

    ret = states.pkg.installed(name=package, refresh=False)
    assert ret.result is True

    ret = modules.pkg.info_installed(package)
    assert pkgquery in str(ret)


@pytest.mark.requires_salt_states("pkg.latest", "pkg.removed")
@pytest.mark.slow_test
def test_pkg_010_latest(PKG_TARGETS, latest_version, states):
    """
    This tests pkg.latest with a package that has no epoch (or a zero
    epoch).
    """
    target = PKG_TARGETS[0]
    version = latest_version(target)

    # If this assert fails, we need to find new targets, this test needs to
    # be able to test successful installation of packages, so this package
    # needs to not be installed before we run the states below
    assert version

    ret = states.pkg.latest(name=target, refresh=False)
    assert ret.result is True
    ret = states.pkg.removed(name=target)
    assert ret.result is True


@pytest.mark.requires_salt_modules("pkg.list_pkgs", "pkg.list_upgrades", "pkg.version")
@pytest.mark.requires_salt_states("pkg.latest")
@pytest.mark.slow_test
def test_pkg_011_latest_only_upgrade(
    grains, PKG_TARGETS, latest_version, states, modules
):
    """
    WARNING: This test will pick a package with an available upgrade (if
    there is one) and upgrade it to the latest version.
    """
    if grains["os_family"] != "Debian":
        pytest.skip("Only runs on Debian based operating systems.")

    target = PKG_TARGETS[0]

    # If this assert fails, we need to find new targets, this test needs to
    # be able to test that the state fails when you try to run the state
    # with only_upgrade=True on a package which is not already installed,
    # so the targeted package needs to not be installed before we run the
    # state below.
    version = latest_version(target)
    assert version

    ret = states.pkg.latest(name=target, refresh=False, only_upgrade=True)
    assert ret.result is False

    # Now look for updates and try to run the state on a package which is already up-to-date.
    installed_pkgs = modules.pkg.list_pkgs()
    updates = modules.pkg.list_upgrades(refresh=False)

    for pkgname in updates:
        if pkgname in installed_pkgs:
            target = pkgname
            break
    else:
        target = ""
        log.warning(
            "No available upgrades to installed packages, skipping "
            "only_upgrade=True test with already-installed package. For "
            "best results run this test on a machine with upgrades "
            "available."
        )

    if target:
        ret = states.pkg.latest(name=target, refresh=False, only_upgrade=True)
        assert ret.result is True
        new_version = modules.pkg.version(target, use_context=False)
        assert new_version == updates[target]
        ret = states.pkg.latest(name=target, refresh=False, only_upgrade=True)
        assert (
            ret.raw["pkg_|-{0}_|-{0}_|-latest".format(target)]["comment"]
            == f"Package {target} is already up-to-date"
        )


@pytest.mark.usefixtures("WILDCARDS_SUPPORTED")
@pytest.mark.requires_salt_modules("pkg.version")
@pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
@pytest.mark.slow_test
def test_pkg_012_installed_with_wildcard_version(PKG_TARGETS, states, modules):
    """
    This is a destructive test as it installs and then removes a package
    """
    target = PKG_TARGETS[0]
    version = modules.pkg.version(target)

    # If this assert fails, we need to find new targets, this test needs to
    # be able to test successful installation of packages, so this package
    # needs to not be installed before we run the states below
    assert not version

    ret = states.pkg.installed(
        name=target,
        version="*",
        refresh=False,
    )
    assert ret.result is True

    # Repeat state, should pass
    ret = states.pkg.installed(
        name=target,
        version="*",
        refresh=False,
    )

    expected_comment = (
        "All specified packages are already installed and are at the desired version"
    )
    assert ret.result is True
    assert ret.raw[next(iter(ret.raw))]["comment"] == expected_comment

    # Repeat one more time with unavailable version, test should fail
    ret = states.pkg.installed(
        name=target,
        version="93413*",
        refresh=False,
    )
    assert ret.result is False

    # Clean up
    ret = states.pkg.removed(name=target)
    assert ret.result is True


@pytest.mark.requires_salt_modules("pkg.version", "pkg.latest_version")
@pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
@pytest.mark.slow_test
def test_pkg_013_installed_with_comparison_operator(
    grains, PKG_TARGETS, states, modules
):
    """
    This is a destructive test as it installs and then removes a package
    """
    if grains["os_family"] != "RedHat" or grains["os_family"] != "Debian":
        pytest.skip("Only runs on Debian or RedHat.")

    target = PKG_TARGETS[0]
    version = modules.pkg.version(target)

    # If this assert fails, we need to find new targets, this test needs to
    # be able to test successful installation of packages, so this package
    # needs to not be installed before we run the states below
    assert not version

    latest_version = modules.pkg.latest_version(target, refresh=False)

    try:
        ret = states.pkg.installed(
            name=target,
            version="<9999999",
            refresh=False,
        )
        assert ret.result is True

        # The version that was installed should be the latest available
        version = modules.pkg.version(target)
        assert version, latest_version
    finally:
        # Clean up
        ret = states.pkg.removed(name=target)
        assert ret.result is True


@pytest.mark.requires_salt_modules("pkg.version")
@pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
@pytest.mark.slow_test
def test_pkg_014_installed_missing_release(grains, PKG_TARGETS, states, modules):
    """
    Tests that a version number missing the release portion still resolves
    as correctly installed. For example, version 2.0.2 instead of 2.0.2-1.el7
    """
    if grains["os_family"] != "RedHat":
        pytest.skip("Only runs on RedHat based operating systems.")

    target = PKG_TARGETS[0]
    version = modules.pkg.version(target)

    # If this assert fails, we need to find new targets, this test needs to
    # be able to test successful installation of packages, so this package
    # needs to not be installed before we run the states below
    assert not version

    ret = states.pkg.installed(
        name=target,
        version=salt.utils.pkg.rpm.version_to_evr(version)[1],
        refresh=False,
    )
    assert ret.result is True

    # Clean up
    ret = states.pkg.removed(name=target)
    assert ret.result is True


@pytest.mark.skip_on_photonos(
    reason="package hold/unhold unsupported on Photon OS",
)
@pytest.mark.requires_salt_modules(
    "pkg.hold", "pkg.unhold", "pkg.version", "pkg.list_pkgs"
)
@pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
@pytest.mark.slow_test
def test_pkg_015_installed_held(grains, modules, states, PKG_TARGETS):
    """
    Tests that a package can be held even when the package is already installed.
    """
    versionlock_pkg = None
    if grains["os_family"] == "RedHat":
        pkgs = {
            p for p in modules.pkg.list_repo_pkgs() if "yum-plugin-versionlock" in p
        }
        if not pkgs:
            pytest.skip("No versionlock package found in repositories")
        for versionlock_pkg in pkgs:
            ret = states.pkg.installed(name=versionlock_pkg, refresh=False)
            # Exit loop if a versionlock package installed correctly
            try:
                assert ret.result is True
                log.debug("Installed versionlock package: %s", versionlock_pkg)
                break
            except AssertionError as exc:
                log.debug("Versionlock package not found:\n%s", exc)
        else:
            pytest.fail(f"Could not install versionlock package from {pkgs}")

    target = PKG_TARGETS[0]

    # First we ensure that the package is installed
    ret = states.pkg.installed(
        name=target,
        refresh=False,
    )
    assert ret.result is True

    # Then we check that the package is now held
    ret = states.pkg.installed(
        name=target,
        hold=True,
        refresh=False,
    )

    if versionlock_pkg and "-versionlock is not installed" in str(ret):
        pytest.skip(f"{ret}  `{versionlock_pkg}` is installed")

    # changes from pkg.hold for Red Hat family are different
    target_changes = {}
    if (
        grains["os_family"] == "RedHat"
        or grains["os"] == "FreeBSD"
        or grains["os_family"] == "Suse"
    ):
        target_changes = {"new": "hold", "old": ""}
    elif grains["os_family"] == "Debian":
        target_changes = {"new": "hold", "old": "install"}

    try:
        tag = "pkg_|-{0}_|-{0}_|-installed".format(target)
        assert ret.result is True
        assert tag in ret.raw
        assert "changes" in ret.raw[tag]
        assert target in ret.raw[tag]["changes"]
        if not target_changes:
            pytest.skip(
                "Test needs to be configured for {}: {}".format(
                    grains["os"], ret.raw[tag]["changes"][target]
                )
            )
        assert ret.raw[tag]["changes"][target] == target_changes
    finally:
        # Clean up, unhold package and remove
        modules.pkg.unhold(name=target)
        ret = states.pkg.removed(name=target)
        assert ret.result is True
        if versionlock_pkg:
            ret = states.pkg.removed(name=versionlock_pkg)
            assert ret.result is True


@pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
@pytest.mark.slow_test
def test_pkg_016_conditionally_ignore_epoch(PKG_EPOCH_TARGETS, latest_version, states):
    """
    See
    https://github.com/saltstack/salt/issues/56654#issuecomment-615034952

    This is a destructive test as it installs a package
    """
    target = PKG_EPOCH_TARGETS[0]

    # Strip the epoch from the latest available version
    version = latest_version(target).split(":", 1)[-1]
    # If this assert fails, we need to find a new target. This test
    # needs to be able to test successful installation of the package, so
    # the target needs to not be installed before we run the
    # pkg.installed state below
    assert version

    # CASE 1: package name passed in "name" param
    ret = states.pkg.installed(name=target, version=version, refresh=False)
    assert ret.result is True
    ret = states.pkg.removed(name=target)
    assert ret.result is True

    # CASE 2: same as case 1 but with "pkgs"
    ret = states.pkg.installed(name="foo", pkgs=[{target: version}], refresh=False)
    assert ret.result is True
    ret = states.pkg.removed(name=target)
    assert ret.result is True


@pytest.mark.skip_on_photonos(
    reason="package hold/unhold unsupported on Photon OS",
)
@pytest.mark.requires_salt_modules(
    "pkg.hold", "pkg.unhold", "pkg.version", "pkg.list_pkgs"
)
@pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
def test_pkg_017_installed_held_equals_false(grains, modules, states, PKG_TARGETS):
    """
    Tests that a package is installed when hold is explicitly False.

    See https://github.com/saltstack/salt/issues/58801.
    """
    versionlock_pkg = None
    if grains["os_family"] == "RedHat":
        from salt.modules.yumpkg import _versionlock_pkg

        pkgs = {
            p for p in modules.pkg.list_repo_pkgs() if _versionlock_pkg(grains) in p
        }
        if not pkgs:
            pytest.skip("No versionlock package found in repositories")
        for versionlock_pkg in pkgs:
            ret = states.pkg.installed(name=versionlock_pkg, refresh=False)
            # Exit loop if a versionlock package installed correctly
            try:
                assert ret.result is True
                log.debug("Installed versionlock package: %s", versionlock_pkg)
                break
            except AssertionError as exc:
                log.debug("Versionlock package not found:\n%s", exc)
        else:
            pytest.fail(f"Could not install versionlock package from {pkgs}")

    target = PKG_TARGETS[0]

    # First we ensure that the package is installed
    target_ret = states.pkg.installed(
        name=target,
        hold=False,
        refresh=False,
    )
    assert target_ret.result is True

    if versionlock_pkg and "-versionlock is not installed" in str(target_ret):
        pytest.skip(f"{target_ret}  `{versionlock_pkg}` is installed")

    try:
        tag = "pkg_|-{0}_|-{0}_|-installed".format(target)
        assert target_ret.result is True
        assert tag in target_ret.raw
        assert "changes" in target_ret.raw[tag]
        # On Centos 7 package is already installed, no change happened
        if target_ret.raw[tag].get("changes"):
            assert target in target_ret.raw[tag]["changes"]
        if grains["os_family"] == "Suse":
            assert "packages were installed" in target_ret.raw[tag]["comment"]
        else:
            #  The "held" string is part of a longer comment that may look
            #  like:
            #
            #    Package units is not being held.
            assert "held" in target_ret.raw[tag]["comment"]
    finally:
        # Clean up, unhold package and remove
        ret = states.pkg.removed(name=target)
        assert ret.result is True
        if versionlock_pkg:
            ret = states.pkg.removed(name=versionlock_pkg)
            assert ret.result is True


@pytest.mark.requires_salt_modules("pkg.version")
@pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
@pytest.mark.slow_test
def test_pkg_cap_001_installed(PKG_CAP_TARGETS, modules, states):
    """
    This is a destructive test as it installs and then removes a package
    """
    target, realpkg = PKG_CAP_TARGETS[0]
    version = modules.pkg.version(target)
    realver = modules.pkg.version(realpkg)

    # If this condition is False, we need to find new targets.
    # This needs to be able to test successful installation of packages.
    # These packages need to not be installed before we run the states below
    if not (version and realver):
        pytest.skip("TODO: New pkg cap targets required")

    try:
        ret = states.pkg.installed(
            name=target,
            refresh=False,
            resolve_capabilities=True,
            test=True,
        )
        assert (
            f"The following packages would be installed/updated: {realpkg}"
            in ret.comment
        )
        ret = states.pkg.installed(
            name=target, refresh=False, resolve_capabilities=True
        )
        assert ret.result is True
    finally:
        ret = states.pkg.removed(name=realpkg)
        assert ret.result is True


@pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
@pytest.mark.slow_test
def test_pkg_cap_002_already_installed(PKG_CAP_TARGETS, modules, states):
    """
    This is a destructive test as it installs and then removes a package
    """
    target, realpkg = PKG_CAP_TARGETS[0]
    version = modules.pkg.version(target)
    realver = modules.pkg.version(realpkg)

    # If this condition is False, we need to find new targets.
    # This needs to be able to test successful installation of packages.
    # These packages need to not be installed before we run the states below
    if not (version and realver):
        pytest.skip("TODO: New pkg cap targets required")

    try:
        # install the package
        ret = states.pkg.installed(name=realpkg, refresh=False)
        assert ret.result is True

        # Try to install again. Nothing should be installed this time.
        ret = states.pkg.installed(
            name=target,
            refresh=False,
            resolve_capabilities=True,
            test=True,
        )
        assert "All specified packages are already installed" in ret.comment

        ret = states.pkg.installed(
            name=target, refresh=False, resolve_capabilities=True
        )
        assert ret.result is True

        assert "packages are already installed" in ret.comment
    finally:
        ret = states.pkg.removed(name=realpkg)
        assert ret.result is True


@pytest.mark.usefixtures("VERSION_SPEC_SUPPORTED")
@pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
@pytest.mark.slow_test
def test_pkg_cap_003_installed_multipkg_with_version(
    PKG_CAP_TARGETS,
    PKG_TARGETS,
    latest_version,
    modules,
    states,
    grains,
):
    """
    This is a destructive test as it installs and then removes two packages
    """
    target, realpkg = PKG_CAP_TARGETS[0]
    version = modules.pkg.version(target)
    realver = modules.pkg.version(realpkg)

    # If this condition is False, we need to find new targets.
    # This needs to be able to test successful installation of packages.
    # These packages need to not be installed before we run the states below
    if not (version and realver):
        pytest.skip("TODO: New pkg cap targets required")

    cleanup_pkgs = PKG_TARGETS
    try:
        pkgs = [
            {PKG_TARGETS[0]: version},
            PKG_TARGETS[1],
            {target: realver},
        ]
        ret = states.pkg.installed(
            name="test_pkg_cap_003_installed_multipkg_with_version-install",
            pkgs=pkgs,
            refresh=False,
        )
        assert ret.result is False

        ret = states.pkg.installed(
            name="test_pkg_cap_003_installed_multipkg_with_version-install-capability",
            pkgs=pkgs,
            refresh=False,
            resolve_capabilities=True,
            test=True,
        )
        assert "packages would be installed/updated" in ret.comment
        assert f"{realpkg}={realver}" in ret.comment

        ret = states.pkg.installed(
            name="test_pkg_cap_003_installed_multipkg_with_version-install-capability",
            pkgs=pkgs,
            refresh=False,
            resolve_capabilities=True,
        )
        assert ret.result is True
        cleanup_pkgs.append(realpkg)
    finally:
        ret = states.pkg.removed(
            name="test_pkg_cap_003_installed_multipkg_with_version-remove",
            pkgs=cleanup_pkgs,
        )
        assert ret.result is True


@pytest.mark.requires_salt_modules("pkg.version")
@pytest.mark.requires_salt_states("pkg.latest", "pkg.removed")
@pytest.mark.slow_test
def test_pkg_cap_004_latest(PKG_CAP_TARGETS, modules, states):
    """
    This tests pkg.latest with a package that has no epoch (or a zero
    epoch).
    """
    target, realpkg = PKG_CAP_TARGETS[0]
    version = modules.pkg.version(target)
    realver = modules.pkg.version(realpkg)

    # If this condition is False, we need to find new targets.
    # This needs to be able to test successful installation of packages.
    # These packages need to not be installed before we run the states below
    if not (version and realver):
        pytest.skip("TODO: New pkg cap targets required")

    try:
        ret = states.pkg.latest(
            name=target,
            refresh=False,
            resolve_capabilities=True,
            test=True,
        )
        assert (
            f"The following packages would be installed/upgraded: {realpkg}"
            in ret.comment
        )
        ret = states.pkg.latest(name=target, refresh=False, resolve_capabilities=True)
        assert ret.result is True

        ret = states.pkg.latest(name=target, refresh=False, resolve_capabilities=True)
        assert ret.result is True
        assert "is already up-to-date" in ret.comment
    finally:
        ret = states.pkg.removed(name=realpkg)
        assert ret.result is True


@pytest.mark.requires_salt_modules("pkg.version")
@pytest.mark.requires_salt_states("pkg.installed", "pkg.removed", "pkg.downloaded")
@pytest.mark.slow_test
def test_pkg_cap_005_downloaded(PKG_CAP_TARGETS, modules, states):
    """
    This is a destructive test as it installs and then removes a package
    """
    target, realpkg = PKG_CAP_TARGETS[0]
    version = modules.pkg.version(target)
    realver = modules.pkg.version(realpkg)

    # If this condition is False, we need to find new targets.
    # This needs to be able to test successful installation of packages.
    # These packages need to not be installed before we run the states below
    if not (version and realver):
        pytest.skip("TODO: New pkg cap targets required")

    ret = states.pkg.downloaded(name=target, refresh=False)
    assert ret.result is False

    ret = states.pkg.downloaded(
        name=target,
        refresh=False,
        resolve_capabilities=True,
        test=True,
    )
    assert f"The following packages would be downloaded: {realpkg}" in ret.comment

    ret = states.pkg.downloaded(name=target, refresh=False, resolve_capabilities=True)
    assert ret.result is True


@pytest.mark.requires_salt_modules("pkg.version")
@pytest.mark.requires_salt_states("pkg.installed", "pkg.removed", "pkg.uptodate")
@pytest.mark.slow_test
def test_pkg_cap_006_uptodate(PKG_CAP_TARGETS, modules, states):
    """
    This is a destructive test as it installs and then removes a package
    """
    target, realpkg = PKG_CAP_TARGETS[0]
    version = modules.pkg.version(target)
    realver = modules.pkg.version(realpkg)

    # If this condition is False, we need to find new targets.
    # This needs to be able to test successful installation of packages.
    # These packages need to not be installed before we run the states below
    if not (version and realver):
        pytest.skip("TODO: New pkg cap targets required")

    try:
        ret = states.pkg.installed(
            name=target, refresh=False, resolve_capabilities=True
        )
        assert ret.result is True
        ret = states.pkg.uptodate(
            name="test_pkg_cap_006_uptodate",
            pkgs=[target],
            refresh=False,
            resolve_capabilities=True,
        )
        assert ret.result is True
        assert "System is already up-to-date" in ret.comment
    finally:
        ret = states.pkg.removed(name=realpkg)
        assert ret.result is True


@pytest.mark.requires_salt_modules(
    "pkg.version", "pkg.latest_version", "pkg.remove", "pkg.purge", "pkg.list_pkgs"
)
@pytest.mark.requires_salt_states("pkg.installed", "pkg.removed", "pkg.purged")
def test_pkg_purged_with_removed_pkg(grains, PKG_TARGETS, states, modules):
    """
    This is a destructive test as it installs and then removes a package, then purges a removed package
    """
    if grains["os_family"] != "Debian":
        pytest.skip("Only runs on Debian.")

    target = PKG_TARGETS[0]

    ret = states.pkg.installed(
        name=target,
        version="<9999999",
        refresh=False,
    )
    assert ret.result is True

    # The version that was installed should be the latest available
    version = modules.pkg.version(target)
    assert version

    # Clean up
    ret = states.pkg.removed(name=target)
    assert ret.result is True

    ret = states.pkg.purged(name=target)
    assert ret.result is True
    assert ret.name == target
    assert ret.comment == "All targeted packages were purged."
    assert ret.changes == {
        "installed": {},
        "removed": {target: {"new": "", "old": version}},
    }


@pytest.mark.skip_unless_on_windows()
def test_pkg_removed_with_version_multiple(install_7zip, modules, states):
    """
    This tests removing a specific version of a package when multiple versions
    are installed. This is specific to Windows. The only version I could find
    that allowed multiple installs of differing versions was 7zip, so we'll use
    that.
    """
    ret = states.pkg.removed(name="7zip", version="19.00.00.0")
    assert ret.result is True
    current = modules.pkg.version("7zip")
    assert "22.01.00.0" in current


@pytest.mark.skip_unless_on_windows()
def test_pkg_latest_test_true(states, modules, state_tree, pkg_def_contents):
    repo_dir = state_tree / "winrepo_ng"
    with pytest.helpers.temp_file("my-software.sls", pkg_def_contents, repo_dir):
        modules.pkg.refresh_db()
    assert len(modules.pkg.get_package_info("my-software")) == 2
    result = states.pkg.latest("my-software", test=True)
    expected = {"my-software": {"new": "1.0.2", "old": ""}}
    assert result.changes == expected



import ast
import copy
import os
import re
import shutil
import textwrap

import pytest
from saltfactories.utils import random_string

import salt.utils.files
import salt.utils.platform
import salt.utils.pycrypto
import salt.utils.yaml

pytestmark = [
    pytest.mark.core_test,
    pytest.mark.windows_whitelisted,
]

PUB_KEY = textwrap.dedent(
    """\
        -----BEGIN PUBLIC KEY-----
        MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAoqIZDtcQtqUNs0wC7qQz
        JwFhXAVNT5C8M8zhI+pFtF/63KoN5k1WwAqP2j3LquTG68WpxcBwLtKfd7FVA/Kr
        OF3kXDWFnDi+HDchW2lJObgfzLckWNRFaF8SBvFM2dys3CGSgCV0S/qxnRAjrJQb
        B3uQwtZ64ncJAlkYpArv3GwsfRJ5UUQnYPDEJwGzMskZ0pHd60WwM1gMlfYmNX5O
        RBEjybyNpYDzpda6e6Ypsn6ePGLkP/tuwUf+q9wpbRE3ZwqERC2XRPux+HX2rGP+
        mkzpmuHkyi2wV33A9pDfMgRHdln2CLX0KgfRGixUQhW1o+Kmfv2rq4sGwpCgLbTh
        NwIDAQAB
        -----END PUBLIC KEY-----
        """
)


def test_remove_key(salt_master, salt_key_cli):
    """
    test salt-key -d usage
    """
    min_name = random_string("minibar-")
    pki_dir = salt_master.config["pki_dir"]
    key = os.path.join(pki_dir, "minions", min_name)

    with salt.utils.files.fopen(key, "w") as fp:
        fp.write(PUB_KEY)

    try:
        # Check Key
        ret = salt_key_cli.run("-p", min_name)
        assert ret.returncode == 0
        assert "minions" in ret.data
        assert min_name in ret.data["minions"]
        assert "-----BEGIN PUBLIC KEY-----" in ret.data["minions"][min_name]
        # Remove Key
        ret = salt_key_cli.run("-d", min_name, "-y")
        assert ret.returncode == 0
        # We can't load JSON because we print to stdout!
        # >>>>> STDOUT >>>>>
        # The following keys are going to be deleted:
        # {
        #     "minions": [
        #         "minibar"
        #     ]
        # }
        # Key for minion minibar deleted.
        # <<<<< STDOUT <<<<<
        assert "minions" in ret.stdout
        assert min_name in ret.stdout
        # Check Key
        ret = salt_key_cli.run("-p", min_name)
        assert ret.returncode == 0
        assert ret.data == {}
    finally:
        if os.path.exists(key):
            os.unlink(key)


def test_remove_all_keys_with_arg(salt_key_cli):
    """
    test salt-key -D usage with args
    """
    # Remove Key
    ret = salt_key_cli.run("-D", "junk")
    assert "Delete all takes no arguments" in ret.stderr


@pytest.mark.skip_if_not_root
@pytest.mark.destructive_test
@pytest.mark.skip_on_windows(reason="PAM is not supported on Windows")
def test_remove_key_eauth(salt_key_cli, salt_master, salt_eauth_account):
    """
    test salt-key -d usage
    """
    min_name = random_string("minibar-")
    pki_dir = salt_master.config["pki_dir"]
    key = os.path.join(pki_dir, "minions", min_name)

    with salt.utils.files.fopen(key, "w") as fp:
        fp.write(PUB_KEY)

    try:
        # Check Key
        ret = salt_key_cli.run("-p", min_name)
        assert ret.returncode == 0
        assert "minions" in ret.data
        assert min_name in ret.data["minions"]
        assert "-----BEGIN PUBLIC KEY-----" in ret.data["minions"][min_name]
        # Remove Key
        ret = salt_key_cli.run(
            "-d",
            min_name,
            "-y",
            "--eauth",
            "pam",
            "--username",
            salt_eauth_account.username,
            "--password",
            salt_eauth_account.password,
        )
        assert ret.returncode == 0
        # We can't load JSON because we print to stdout!
        # >>>>> STDOUT >>>>>
        # The following keys are going to be deleted:
        # {
        #     "minions": [
        #         "minibar"
        #     ]
        # }
        # Key for minion minibar deleted.
        # <<<<< STDOUT <<<<<
        assert "minions" in ret.stdout
        assert min_name in ret.stdout
        # Check Key
        ret = salt_key_cli.run("-p", min_name)
        assert ret.returncode == 0
        assert ret.data == {}
    finally:
        if os.path.exists(key):
            os.unlink(key)


@pytest.mark.parametrize("key_type", ("acc", "pre", "den", "un", "rej"))
def test_list_accepted_args(salt_key_cli, key_type):
    """
    test salt-key -l for accepted arguments
    """
    # Should not trigger any error
    ret = salt_key_cli.run("-l", key_type)
    assert ret.returncode == 0
    assert "error:" not in ret.stdout
    # Should throw an error now
    ret = salt_key_cli.run("-l", f"foo-{key_type}")
    assert ret.returncode != 0
    assert "error:" in ret.stderr


def test_list_all(salt_key_cli, salt_minion, salt_sub_minion):
    """
    test salt-key -L
    """
    ret = salt_key_cli.run("-L")
    assert ret.returncode == 0
    expected = {
        "minions_rejected": [],
        "minions_denied": [],
        "minions_pre": [],
        "minions": [salt_minion.id, salt_sub_minion.id],
    }
    assert ret.data == expected


def test_list_all_no_check_files(
    salt_key_cli, salt_minion, salt_sub_minion, tmp_path, salt_master
):
    """
    test salt-key -L
    """
    config_dir = tmp_path / "key_no_check_files"
    config_dir.mkdir()
    pki_dir = config_dir / "pki_dir"
    shutil.copytree(salt_master.config["pki_dir"], str(pki_dir))
    with pytest.helpers.change_cwd(str(config_dir)):
        master_config = copy.deepcopy(salt_master.config)
        master_config["pki_check_files"] = False
        master_config["pki_dir"] = "pki_dir"
        master_config["root_dir"] = str(config_dir)
        with salt.utils.files.fopen(str(config_dir / "master"), "w") as fh_:
            fh_.write(salt.utils.yaml.dump(master_config, default_flow_style=False))
        ret = salt_key_cli.run(
            f"--config-dir={config_dir}",
            "-L",
        )
        assert ret.returncode == 0
        expected = {
            "minions_rejected": [],
            "minions_denied": [],
            "minions_pre": [],
            "minions": [salt_minion.id, salt_sub_minion.id],
        }
        assert ret.data == expected

        bad_key = pki_dir / "minions" / "dir1"
        bad_key.mkdir()

        ret = salt_key_cli.run(
            f"--config-dir={config_dir}",
            "-L",
        )
        assert ret.returncode == 0
        assert ret.data == expected


def test_list_all_yaml_out(salt_key_cli, salt_minion, salt_sub_minion):
    """
    test salt-key -L --out=yaml
    """
    ret = salt_key_cli.run("-L", "--out=yaml")
    assert ret.returncode == 0
    output = salt.utils.yaml.safe_load(str(ret.stdout))
    expected = {
        "minions_rejected": [],
        "minions_denied": [],
        "minions_pre": [],
        "minions": [salt_minion.id, salt_sub_minion.id],
    }
    assert output == expected


def test_list_all_raw_out(salt_key_cli, salt_minion, salt_sub_minion):
    """
    test salt-key -L --out=raw
    """
    ret = salt_key_cli.run("-L", "--out=raw")
    assert ret.returncode == 0
    output = ast.literal_eval(ret.stdout)
    expected = {
        "minions_rejected": [],
        "minions_denied": [],
        "minions_pre": [],
        "minions": [salt_minion.id, salt_sub_minion.id],
    }
    assert output == expected


def test_list_acc(salt_key_cli, salt_minion, salt_sub_minion):
    """
    test salt-key -l acc
    """
    ret = salt_key_cli.run("-l", "acc")
    assert ret.returncode == 0
    expected = {"minions": [salt_minion.id, salt_sub_minion.id]}
    assert ret.data == expected


@pytest.mark.skip_if_not_root
@pytest.mark.destructive_test
@pytest.mark.skip_on_windows(reason="PAM is not supported on Windows")
def test_list_acc_eauth(salt_key_cli, salt_minion, salt_sub_minion, salt_eauth_account):
    """
    test salt-key -l with eauth
    """
    ret = salt_key_cli.run(
        "-l",
        "acc",
        "--eauth",
        "pam",
        "--username",
        salt_eauth_account.username,
        "--password",
        salt_eauth_account.password,
    )
    assert ret.returncode == 0
    expected = {"minions": [salt_minion.id, salt_sub_minion.id]}
    assert ret.data == expected


@pytest.mark.skip_if_not_root
@pytest.mark.destructive_test
@pytest.mark.skip_on_windows(reason="PAM is not supported on Windows")
def test_list_acc_eauth_bad_creds(salt_key_cli, salt_eauth_account):
    """
    test salt-key -l with eauth and bad creds
    """
    ret = salt_key_cli.run(
        "-l",
        "acc",
        "--eauth",
        "pam",
        "--username",
        salt_eauth_account.username,
        "--password",
        "wrongpassword",
    )
    assert (
        ret.stdout
        == 'Authentication failure of type "eauth" occurred for user {}.'.format(
            salt_eauth_account.username
        )
    )


def test_list_acc_wrong_eauth(salt_key_cli, salt_eauth_account):
    """
    test salt-key -l with wrong eauth
    """
    ret = salt_key_cli.run(
        "-l",
        "acc",
        "--eauth",
        "wrongeauth",
        "--username",
        salt_eauth_account.username,
        "--password",
        salt_eauth_account.password,
    )
    assert ret.returncode == 0, ret
    assert re.search(
        r"^The specified external authentication system \"wrongeauth\" is not"
        r" available\nAvailable eauth types: auto, .*",
        ret.stdout.replace("\r\n", "\n"),
    )


def test_list_un(salt_key_cli):
    """
    test salt-key -l un
    """
    ret = salt_key_cli.run("-l", "un")
    assert ret.returncode == 0
    expected = {"minions_pre": []}
    assert ret.data == expected


def test_keys_generation(salt_key_cli, tmp_path):
    ret = salt_key_cli.run("--gen-keys", "minibar", "--gen-keys-dir", str(tmp_path))
    assert ret.returncode == 0
    try:
        key_names = ("minibar.pub", "minibar.pem")
        for fname in key_names:
            fpath = tmp_path / fname
            assert fpath.is_file()
    finally:
        for filename in tmp_path.iterdir():
            filename.chmod(0o700)


def test_gen_keys_dir_without_gen_keys(salt_key_cli, tmp_path):
    gen_keys_path = tmp_path / "temp-gen-keys-path"
    ret = salt_key_cli.run("--gen-keys-dir", str(gen_keys_path))
    assert ret.returncode == 0
    assert not gen_keys_path.exists()


def test_keys_generation_keysize_min(salt_key_cli, tmp_path):
    ret = salt_key_cli.run(
        "--gen-keys", "minibar", "--gen-keys-dir", str(tmp_path), "--keysize", "1024"
    )
    assert ret.returncode != 0
    assert "error: The minimum value for keysize is 2048" in ret.stderr


def test_keys_generation_keysize_max(salt_key_cli, tmp_path):
    ret = salt_key_cli.run(
        "--gen-keys", "minibar", "--gen-keys-dir", str(tmp_path), "--keysize", "32769"
    )
    assert ret.returncode != 0
    assert "error: The maximum value for keysize is 32768" in ret.stderr


def test_accept_bad_key(salt_master, salt_key_cli):
    """
    test salt-key -d usage
    """
    min_name = random_string("minibar-")
    pki_dir = salt_master.config["pki_dir"]
    key = os.path.join(pki_dir, "minions_pre", min_name)

    with salt.utils.files.fopen(key, "w") as fp:
        fp.write("")

    try:
        # Check Key
        ret = salt_key_cli.run("-y", "-a", min_name)
        assert ret.returncode == 0
        assert f"invalid key for {min_name}" in ret.stderr
    finally:
        if os.path.exists(key):
            os.remove(key)



"""
    :synopsis: Unit Tests for Windows PKI Module 'state.win_pki'
    :platform: Windows
    :maturity: develop
    .. versionadded:: 2017.7.0
"""

import pytest

import salt.states.win_pki as win_pki
from tests.support.mock import MagicMock, patch

CERT_PATH = r"C:\certs\testdomain.local.cer"
THUMBPRINT = "9988776655443322111000AAABBBCCCDDDEEEFFF"
STORE_PATH = r"Cert:\LocalMachine\My"

CERTS = {
    THUMBPRINT: {
        "dnsnames": ["testdomain.local"],
        "serialnumber": "0123456789AABBCCDD",
        "subject": "CN=testdomain.local, OU=testou, O=testorg, S=California, C=US",
        "thumbprint": THUMBPRINT,
        "version": 3,
    }
}


@pytest.fixture
def configure_loader_modules():
    return {win_pki: {}}


def test_import_cert():
    """
    Test - Import the certificate file into the given certificate store.
    """
    kwargs = {"name": CERT_PATH}
    ret = {
        "name": kwargs["name"],
        "changes": {"old": None, "new": THUMBPRINT},
        "comment": "Certificate '{}' imported into store: {}".format(
            THUMBPRINT, STORE_PATH
        ),
        "result": True,
    }
    mock_cache_file = MagicMock(return_value=CERT_PATH)
    mock_certs = MagicMock(return_value={})
    mock_cert_file = MagicMock(return_value=CERTS[THUMBPRINT])
    mock_import_cert = MagicMock(return_value=True)
    with patch.dict(
        win_pki.__salt__,
        {
            "cp.cache_file": mock_cache_file,
            "win_pki.get_certs": mock_certs,
            "win_pki.get_cert_file": mock_cert_file,
            "win_pki.import_cert": mock_import_cert,
        },
    ):
        with patch.dict(win_pki.__opts__, {"test": False}):
            assert win_pki.import_cert(**kwargs) == ret
        with patch.dict(win_pki.__opts__, {"test": True}):
            ret["comment"] = (
                "Certificate '{}' will be imported into store: {}"
            ).format(THUMBPRINT, STORE_PATH)
            ret["result"] = None
            assert win_pki.import_cert(**kwargs) == ret


def test_remove_cert():
    """
    Test - Remove the certificate from the given certificate store.
    """
    kwargs = {"name": "remove-cert", "thumbprint": THUMBPRINT}
    ret = {
        "name": kwargs["name"],
        "changes": {"old": kwargs["thumbprint"], "new": None},
        "comment": "Certificate '{}' removed from store: {}".format(
            kwargs["thumbprint"], STORE_PATH
        ),
        "result": True,
    }
    mock_certs = MagicMock(return_value=CERTS)
    mock_remove_cert = MagicMock(return_value=True)
    with patch.dict(
        win_pki.__salt__,
        {"win_pki.get_certs": mock_certs, "win_pki.remove_cert": mock_remove_cert},
    ):
        with patch.dict(win_pki.__opts__, {"test": False}):
            assert win_pki.remove_cert(**kwargs) == ret
        with patch.dict(win_pki.__opts__, {"test": True}):
            ret["comment"] = ("Certificate '{}' will be removed from store: {}").format(
                kwargs["thumbprint"], STORE_PATH
            )
            ret["result"] = None
            assert win_pki.remove_cert(**kwargs) == ret



"""
Unit tests for Network runner
"""

import logging

import pytest

import salt.runners.network as network
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def mac_addr_list():
    test_list_mac_addresses = [
        "08:00:27:82:b2:ca",
        "52:54:00:ee:eb:e1",
        "52:54:00:ee:eb:e1",
    ]
    return test_list_mac_addresses.sort()


@pytest.fixture
def id_minion():
    return "test-host"


@pytest.fixture
def cache_grain_data(id_minion):
    return {
        id_minion: {
            "cwd": "/",
            "ip_gw": True,
            "ip4_gw": "192.168.0.1",
            "ip6_gw": False,
            "dns": {
                "nameservers": ["192.168.0.1"],
                "ip4_nameservers": ["192.168.0.1"],
                "ip6_nameservers": [],
                "sortlist": [],
                "domain": "",
                "search": ["example.org"],
                "options": [],
            },
            "fqdns": ["Unknown.example.org"],
            "machine_id": "ae886ddffbcc4f0da1e72769adfe0171",
            "master": "192.168.0.109",
            "server_id": 644891398,
            "localhost": "Unknown.example.org",
            "fqdn": "Unknown.example.org",
            "host": "Unknown",
            "domain": "example.org",
            "hwaddr_interfaces": {
                "lo": "00:00:00:00:00:00",
                "enp0s3": "08:00:27:82:b2:ca",
                "virbr0": "52:54:00:ee:eb:e1",
                "virbr0-nic": "52:54:00:ee:eb:e1",
            },
            "id": "test-host",
            "ip4_interfaces": {
                "lo": ["127.0.0.1"],
                "enp0s3": ["192.168.0.124"],
                "virbr0": ["192.168.122.1"],
                "virbr0-nic": [],
            },
            "ip6_interfaces": {
                "lo": ["::1"],
                "enp0s3": ["fe80::a00:27ff:fe82:b2ca"],
                "virbr0": [],
                "virbr0-nic": [],
            },
            "ipv4": ["127.0.0.1", "192.168.0.124", "192.168.122.1"],
            "ipv6": ["::1", "fe80::a00:27ff:fe82:b2ca"],
            "fqdn_ip4": ["192.168.0.70"],
            "fqdn_ip6": [],
            "ip_interfaces": {
                "lo": ["127.0.0.1", "::1"],
                "enp0s3": ["192.168.0.124", "fe80::a00:27ff:fe82:b2ca"],
                "virbr0": ["192.168.122.1"],
                "virbr0-nic": [],
            },
            "kernelparams": [
                ["BOOT_IMAGE", "/vmlinuz-3.10.0-1127.18.2.el7.x86_64"],
                ["root", "/dev/mapper/centos-root"],
                ["ro", None],
                ["rd.lvm.lv", "centos/root"],
                ["rd.lvm.lv", "centos/swap"],
                ["rhgb", None],
                ["quiet", None],
                ["LANG", "en_US.UTF-8"],
            ],
            "locale_info": {
                "defaultlanguage": "en_US",
                "defaultencoding": "UTF-8",
                "detectedencoding": "UTF-8",
                "timezone": "unknown",
            },
            "num_gpus": 1,
            "gpus": [{"vendor": "vmware", "model": "SVGA II Adapter"}],
            "kernel": "Linux",
            "nodename": "Unknown.example.org",
            "kernelrelease": "3.10.0-1127.18.2.el7.x86_64",
            "kernelversion": "#1 SMP Sun Jul 26 15:27:06 UTC 2020",
            "cpuarch": "x86_64",
            "selinux": {"enabled": False, "enforced": "Disabled"},
            "systemd": {
                "version": "219",
                "features": (
                    "+PAM +AUDIT +SELINUX +IMA -APPARMOR +SMACK +SYSVINIT +UTMP"
                    " +LIBCRYPTSETUP +GCRYPT +GNUTLS +ACL +XZ +LZ4 -SECCOMP +BLKID"
                    " +ELFUTILS +KMOD +IDN"
                ),
            },
            "init": "systemd",
            "lsb_distrib_id": "CentOS Linux",
            "lsb_distrib_codename": "CentOS Linux 7 (Core)",
            "osfullname": "CentOS Linux",
            "osrelease": "7.8.2003",
            "oscodename": "CentOS Linux 7 (Core)",
            "os": "CentOS",
            "num_cpus": 1,
            "cpu_model": "Intel(R) Core(TM) i7-8750H CPU @ 2.20GHz",
            "cpu_flags": [
                "fpu",
                "vme",
                "de",
                "pse",
                "tsc",
                "msr",
                "pae",
                "mce",
                "cx8",
                "apic",
                "sep",
                "mtrr    ",
                "pge",
                "mca",
                "cmov",
                "pat",
                "pse36",
                "clflush",
                "mmx",
                "fxsr",
                "sse",
                "sse2",
                "ht",
                "syscall",
                "nx",
                "rdtscp",
                "lm",
                "constant_tsc",
                "rep_good",
                "nopl",
                "xtopology",
                "nonstop_tsc",
                "eagerfpu",
                "pni",
                "pclmulqdq",
                "monitor",
                "ssse3",
                "cx16",
                "pcid",
                "sse4_1",
                "sse4_2",
                "x2apic",
                "movbe",
                "popcnt",
                "aes",
                "xsave",
                "avx",
                "rdrand",
                "hypervisor",
                "lahf_lm",
                "abm",
                "3dnowprefetch",
                "invpcid_single",
                "fsgsbase",
                "avx2",
                "inv    pcid",
                "rdseed",
                "clflushopt",
                "md_clear",
                "flush_l1d",
            ],
            "os_family": "RedHat",
            "osarch": "x86_64",
            "mem_total": 1998,
            "swap_total": 2047,
            "biosversion": "VirtualBox",
            "productname": "VirtualBox",
            "manufacturer": "innotek GmbH",
            "biosreleasedate": "12/01/2006",
            "uuid": "dd95fedd-1a2b-5e48-86a7-7e339f9f02a1",
            "serialnumber": "0",
            "virtual": "VirtualBox",
            "ps": "ps -efHww",
            "osrelease_info": [7, 8, 2003],
            "osmajorrelease": 7,
            "osfinger": "CentOS Linux-7",
            "path": "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin",
            "systempath": [
                "/usr/local/sbin",
                "/usr/local/bin",
                "/usr/sbin",
                "/usr/bin",
            ],
            "pythonexecutable": "/usr/bin/python3",
            "pythonpath": [
                "/usr/bin",
                "/usr/lib64/python36.zip",
                "/usr/lib64/python3.6",
                "/usr/lib64/python3.6/lib-dynload",
                "/usr/lib64/python3.6/site-packages",
                "/usr/lib/python3.6/site-packages",
            ],
            "pythonversion": [3, 6, 8, "final", 0],
            "saltpath": "/usr/lib/python3.6/site-packages/salt",
            "saltversion": "3003",
            "saltversioninfo": [3003],
            "zmqversion": "4.1.4",
            "disks": ["sda", "sr0"],
            "ssds": [],
            "shell": "/bin/sh",
            "lvm": {"centos": ["root", "swap"]},
            "mdadm": [],
            "username": "root",
            "groupname": "root",
            "pid": 2469,
            "gid": 0,
            "uid": 0,
            "zfs_support": False,
            "zfs_feature_flags": False,
        }
    }


@pytest.fixture
def configure_loader_modules():

    return {
        network: {
            "__grains__": {
                "osarch": "x86_64",
                "os_family": "Redhat",
                "osmajorrelease": 7,
                "kernelrelease": "3.10.0-1127.18.2.el7.x86_64",
            },
        },
    }


def test_wolmatch(cache_grain_data, id_minion, mac_addr_list):
    """
    Test wolmatch
    """
    cache_mock = MagicMock(return_value=cache_grain_data)
    patches = {
        "cache.grains": cache_mock,
    }
    wol_out = MagicMock(return_value=mac_addr_list)
    with patch.dict(network.__salt__, patches):
        with patch("salt.runners.network.wol", wol_out):
            added = network.wolmatch(id_minion)
            assert added.sort() == mac_addr_list



"""
Tests for file.rename function
"""

import os
import shutil

import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]


@pytest.fixture(scope="module")
def file(modules):
    return modules.file


@pytest.fixture(scope="module")
def multiline_string():
    return """\
        Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nam rhoncus
        enim ac bibendum vulputate. Etiam nibh velit, placerat ac auctor in,
        lacinia a turpis. Nulla elit elit, ornare in sodales eu, aliquam sit
        amet nisl.

        Fusce ac vehicula lectus. Vivamus justo nunc, pulvinar in ornare nec,
        sollicitudin id sem. Pellentesque sed ipsum dapibus, dapibus elit id,
        malesuada nisi.

        Lorem ipsum dolor sit amet, consectetur adipiscing elit. Donec
        venenatis tellus eget massa facilisis, in auctor ante aliquet. Sed nec
        cursus metus. Curabitur massa urna, vehicula id porttitor sed, lobortis
        quis leo.
        """


@pytest.fixture(scope="function")
def multiline_file(tmp_path_factory, multiline_string):
    temp_dir = tmp_path_factory.mktemp("replace-tests")
    test_file = temp_dir / "multiline-file.txt"
    test_file.write_text(multiline_string)
    yield test_file
    shutil.rmtree(str(temp_dir))


def test_no_backup(file, multiline_file):
    # Backup file should NOT be created
    bak_file = f"{multiline_file}.bak"
    assert "Salticus" not in multiline_file.read_text()
    file.replace(str(multiline_file), "Etiam", "Salticus", backup=False)
    assert "Salticus" in multiline_file.read_text()
    assert not os.path.exists(bak_file)


def test_backup(file, multiline_file):
    # Should create a backup file. This is basically the default
    bak_file = f"{multiline_file}.bak"
    file.replace(str(multiline_file), "Etiam", "Salticus")
    assert "Salticus" in multiline_file.read_text()
    assert os.path.exists(bak_file)


def test_append_if_not_found_no_match_newline(file):
    contents = "foo=1\nbar=2\n"
    expected = "foo=1\nbar=2\nbaz=\\g<value>\n"
    with pytest.helpers.temp_file("test_file.txt", contents) as target:
        file.replace(
            path=str(target),
            pattern="#*baz=(?P<value>.*)",
            repl="baz=\\g<value>",
            append_if_not_found=True,
        )
        assert target.read_text() == expected


def test_append_if_not_found_no_match_no_newline(file):
    contents = "foo=1\nbar=2"
    expected = "foo=1\nbar=2\nbaz=\\g<value>\n"
    with pytest.helpers.temp_file("test_file.txt", contents) as target:
        file.replace(
            path=str(target),
            pattern="#*baz=(?P<value>.*)",
            repl="baz=\\g<value>",
            append_if_not_found=True,
        )
        assert target.read_text() == expected


def test_append_if_not_found_empty_file(file):
    # A newline should NOT be added in empty files
    contents = None
    expected = "baz=\\g<value>\n"
    with pytest.helpers.temp_file("test_file.txt", contents) as target:
        file.replace(
            path=str(target),
            pattern="#*baz=(?P<value>.*)",
            repl="baz=\\g<value>",
            append_if_not_found=True,
        )
        assert target.read_text() == expected


def test_append_if_not_found_content(file):
    # Using not_found_content, rather than repl
    contents = None
    expected = "baz=3\n"
    with pytest.helpers.temp_file("test_file.txt", contents) as target:
        file.replace(
            path=str(target),
            pattern="#*baz=(?P<value>.*)",
            repl="baz=\\g<value>",
            append_if_not_found=True,
            not_found_content="baz=3",
        )
        assert target.read_text() == expected


def test_append_if_not_found_no_append_on_match(file):
    # Not appending if matches
    contents = "foo=1\nbaz=42\nbar=2"
    with pytest.helpers.temp_file("test_file.txt", contents) as target:
        file.replace(
            path=str(target),
            pattern="#*baz=(?P<value>.*)",
            repl="baz=\\g<value>",
            append_if_not_found=True,
            not_found_content="baz=3",
        )
        assert target.read_text() == contents


def test_dry_run(file, multiline_file):
    before_time = os.stat(str(multiline_file)).st_mtime
    file.replace(str(multiline_file), r"Etiam", "Salticus", dry_run=True)
    after_time = os.stat(str(multiline_file)).st_mtime
    assert before_time == after_time


def test_show_changes(file, multiline_file):
    ret = file.replace(str(multiline_file), r"Etiam", "Salticus", show_changes=True)
    assert ret.startswith("---")  # looks like a diff


def test_no_show_changes(file, multiline_file):
    ret = file.replace(str(multiline_file), r"Etiam", "Salticus", show_changes=False)
    assert isinstance(ret, bool)


def test_re_str_flags(file, multiline_file):
    file.replace(
        str(multiline_file), r"etiam", "Salticus", flags=["MULTILINE", "ignorecase"]
    )
    assert "Salticus" in multiline_file.read_text()


def test_re_int_flags(file, multiline_file):
    # flag for multiline and ignore case is 10
    file.replace(str(multiline_file), r"etiam", "Salticus", flags=10)
    assert "Salticus" in multiline_file.read_text()


def test_numeric_repl(file, multiline_file):
    """
    This test covers cases where the replacement string is numeric. The CLI
    parser yaml-fies it into a numeric type. If not converted back to a string
    type in file.replace, a TypeError occurs when the replace is attempted. See
    https://github.com/saltstack/salt/issues/9097 for more information.
    """
    file.replace(str(multiline_file), r"Etiam", 123)
    assert "123" in multiline_file.read_text()


def test_search_only_return_true(file, multiline_file):
    ret = file.replace(str(multiline_file), r"Etiam", "Salticus", search_only=True)
    assert isinstance(ret, bool)
    assert ret is True


def test_search_only_return_false(file, multiline_file):
    ret = file.replace(str(multiline_file), r"Etian", "Salticus", search_only=True)
    assert isinstance(ret, bool)
    assert ret is False


def test_symlink(file, multiline_file):
    # https://github.com/saltstack/salt/pull/61326
    try:
        # Create a symlink to target
        sym_link = multiline_file.parent / "symlink.lnk"
        sym_link.symlink_to(multiline_file)
        # file.replace on the symlink
        file.replace(str(sym_link), r"Etiam", "Salticus")
        # test that the target was changed
        assert "Salticus" in multiline_file.read_text()
    finally:
        if os.path.exists(str(sym_link)):
            sym_link.unlink()


def test_replace_no_modify_time_update_on_no_change(file, multiline_file):
    os.utime(str(multiline_file), (1552661253, 1552661253))
    mtime = os.stat(str(multiline_file)).st_mtime
    file.replace(str(multiline_file), r"Etia.", "Etiam", backup=False)
    nmtime = os.stat(str(multiline_file)).st_mtime

    assert mtime == nmtime


def test_backslash_literal(file, multiline_file):
    file.replace(str(multiline_file), r"Etiam", "Emma", backslash_literal=True)
    assert "Emma" in multiline_file.read_text()
